import random
from typing import Union, Optional

import torch

from .utils import load_audio
from ..audio.utils import resample
from ..default import cached_model_instances


def is_demucs_available():
    from importlib.util import find_spec
    if find_spec('demucs') is None:
        raise ModuleNotFoundError("Please install Demucs; "
                                  "'pip install -U demucs' or "
                                  "'pip install -U git+https://github.com/facebookresearch/demucs#egg=demucs'; "
                                  "Official Demucs repo: https://github.com/facebookresearch/demucs")


def load_demucs_model(cache: bool = True):
    model_name = 'htdemucs'
    _model_cache = cached_model_instances['demucs'] if cache else None
    if _model_cache is not None and _model_cache[model_name] is not None:
        return _model_cache[model_name]
    is_demucs_available()
    from demucs.pretrained import get_model_from_args
    model = get_model_from_args(type('args', (object,), dict(name=model_name, repo=None))).cpu().eval().models[0]
    if _model_cache is not None:
        _model_cache[model_name] = model
    return model


def apply_demucs_model(
        model,
        mix,
        shifts=0,
        split=True,
        overlap=0.25,
        transition_power=1.,
        progress=False,
        device=None,
        num_workers=0,
        pool=None
):
    if device is None:
        device = mix.device
    else:
        device = torch.device(device)
        mix = mix.to(device)
    if pool is None:
        if num_workers > 0 and device.type == 'cpu':
            from concurrent.futures import ThreadPoolExecutor
            pool = ThreadPoolExecutor(num_workers)
        else:
            from demucs.utils import DummyPoolExecutor
            pool = DummyPoolExecutor()

    from demucs.apply import TensorChunk, tensor_chunk
    from demucs.utils import center_trim

    model = model.to(device)
    assert transition_power >= 1, "transition_power < 1 leads to weird behavior."

    max_shift = int(0.5 * model.samplerate)

    def _inference(_mix):
        _length = _mix.shape[-1]
        valid_length = model.valid_length(_length) if hasattr(model, 'valid_length') else _length
        padded_mix = tensor_chunk(_mix).padded(valid_length).to(device)
        with torch.no_grad():
            out = model(padded_mix)
        return center_trim(out, _length)

    def _split(_mix):
        batch, channels, length = _mix.shape
        out = torch.zeros(batch, len(model.sources), channels, length, device=device)
        sum_weight = torch.zeros(length, device=device)
        segment = int(model.samplerate * model.segment)
        stride = int((1 - overlap) * segment)
        offsets = range(0, length, stride)
        # We start from a triangle shaped weight, with maximal weight in the middle
        # of the segment. Then we normalize and take to the power `transition_power`.
        # Large values of transition power will lead to sharper transitions.
        weight = torch.cat(
            [torch.arange(1, segment // 2 + 1, device=device),
             torch.arange(segment - segment // 2, 0, -1, device=device)]
        )
        assert len(weight) == segment
        # If the overlap < 50%, this will translate to linear transition when
        # transition_power is 1.
        weight = (weight / weight.max()) ** transition_power
        futures = []
        for offset in offsets:
            chunk = TensorChunk(_mix, offset, segment)
            future = pool.submit(_inference, chunk)
            futures.append((future, offset))
            offset += segment
        samples_per_future = length / len(futures)
        for future, offset in futures:
            chunk_out = future.result()
            chunk_length = chunk_out.shape[-1]
            out[..., offset:offset + segment] += (weight[:chunk_length] * chunk_out).to(mix.device)
            sum_weight[offset:offset + segment] += weight[:chunk_length].to(mix.device)
            update_pbar(samples_per_future)
        assert sum_weight.min() > 0
        out /= sum_weight
        return out

    orig_length = mix.shape[-1]
    if progress:
        import tqdm
        total_duration = round(orig_length / model.samplerate, 2)
        pbar = tqdm.tqdm(total=total_duration, unit='sec', desc='Demucs')

        def update_pbar(samples):
            if samples is None:
                pbar.update(pbar.total - pbar.n)
                return
            if shifts > 1:
                samples /= shifts

            seek_duration = min(round(pbar.n + (samples / model.samplerate), 2), total_duration)
            pbar.update(seek_duration - pbar.n)  # this keeps ``n`` rounded

    else:
        def update_pbar(samples):
            pass

    inference = _split if split else _inference

    if not shifts:
        output = inference(mix)
    else:
        output = 0
        mix = tensor_chunk(mix).padded(orig_length + 2 * max_shift)
        for _ in range(shifts):
            shift_offset = random.randint(0, max_shift)
            shifted = TensorChunk(mix, shift_offset, orig_length + max_shift - shift_offset)
            shifted_out = inference(shifted)
            output += shifted_out[..., max_shift - shift_offset:]
        output /= shifts

    update_pbar(None)

    return output[0, model.sources.index('vocals')].mean(0)


def demucs_audio(
        audio: Union[torch.Tensor, str, bytes],
        input_sr: int = None,
        output_sr: int = None,
        model=None,
        device=None,
        verbose: bool = True,
        save_path: Optional[Union[str, callable]] = None,
        seed: Optional[int] = 1,
        **demucs_options
) -> torch.Tensor:
    """
    Isolates vocals / remove noise from ``audio`` with Demucs.

    Official repo, https://github.com/facebookresearch/demucs.
    """
    if model is None:
        model = load_demucs_model()

    if isinstance(audio, (str, bytes)):
        audio = torch.from_numpy(load_audio(audio, model.samplerate))
    elif input_sr != model.samplerate:
        if input_sr is None:
            raise ValueError('No ``input_sr`` specified for audio tensor.')
        audio = resample(audio, input_sr, model.samplerate)
    audio_dims = audio.dim()
    assert audio_dims <= 3
    if dims_missing := 3 - audio_dims:
        audio = audio[[None]*dims_missing]
    if audio.shape[-2] == 1:
        audio = audio.repeat_interleave(2, -2)

    if device is None:
        device = "cuda" if torch.cuda.is_available() else "cpu"

    apply_kwarg = dict(
        model=model,
        mix=audio,
        device=device,
        split=True,
        overlap=.25,
        progress=verbose is not None,
    )
    apply_kwarg.update(demucs_options)
    if seed is not None:
        random.seed(seed)
    vocals = apply_demucs_model(**apply_kwarg)

    if device != 'cpu':
        torch.cuda.empty_cache()

    if output_sr is not None and model.samplerate != output_sr:
        vocals = resample(vocals, model.samplerate, output_sr)

    if save_path is not None:
        if isinstance(save_path, str):
            from .output import save_audio_tensor
            save_audio_tensor(vocals, save_path, output_sr or model.samplerate, verbose=verbose)
        else:
            save_path(vocals)

    return vocals



import subprocess
import warnings
from typing import Union, Optional, BinaryIO, Tuple, Iterable

import numpy as np
import torch
import torchaudio

from whisper.audio import SAMPLE_RATE


def is_ytdlp_available():
    return subprocess.run('yt-dlp -h', shell=True, capture_output=True).returncode == 0


def load_source(
        source: Union[str, bytes],
        verbose: Optional[bool] = False,
        only_ffmpeg: bool = False,
        return_dict: bool = False
) -> Union[bytes, dict]:
    """
    Return content downloaded by YT-DLP if ``source`` is a URL else return ``source``.
    """
    if not only_ffmpeg and isinstance(source, str) and '://' in source:
        if is_ytdlp_available():
            if return_dict:
                stderr = subprocess.PIPE
                verbosity = ' --no-simulate --print title,duration,is_live'
            else:
                verbosity = ' -q' if verbose is None else (' --progress' if verbose else ' --progress -q')
                stderr = None
            p = subprocess.Popen(
                f'yt-dlp "{source}" -f ba/w -I 1{verbosity} -o -',
                shell=True,
                stderr=stderr,
                stdout=subprocess.PIPE,
                bufsize=0
            )
            if return_dict:
                title = p.stderr.readline().decode('utf-8', errors='ignore').strip('\n') or None
                duration = p.stderr.readline().decode('utf-8', errors='ignore').strip('\n')
                is_live = p.stderr.readline().decode('utf-8', errors='ignore').strip('\n')
                try:
                    duration = int(duration)
                except ValueError:
                    duration = None
                is_live = (True if is_live == 'True' else False) if is_live in ('True', 'False') else None
                if verbose is not None:
                    print(f'Media Info (YT-DLP):\n'
                          f'-Title: "{title or "N/A"}"\n'
                          f'-Duration: {duration}s\n'
                          f'-Live: {is_live}')
                return dict(popen=p, title=title, duration=duration, is_live=is_live)
            return p.communicate()[0]
        else:
            warnings.warn('URL detected but yt-dlp not available. '
                          'To handle a greater variety of URLs (i.e. non-direct links), '
                          'install yt-dlp, \'pip install yt-dlp\' (repo: https://github.com/yt-dlp/yt-dlp).')
    return source


def load_audio(
        file: Union[str, bytes, BinaryIO],
        sr: int = None,
        verbose: Optional[bool] = True,
        only_ffmpeg: bool = False,

):
    """
    Open an audio file and read as mono waveform then resamples as necessary.

    Parameters
    ----------
    file : str or bytes or BinaryIO
        The audio file to open, bytes of file, or URL to audio/video.
    sr : int, default whisper.model.SAMPLE_RATE
        The sample rate to resample the audio if necessary.
    verbose : bool or None, default True
        Verbosity for yd-dlp and displaying content metadata when ``file`` is a URL. If not ``None``, display metadata.
        For yd-dlp: ``None`` is "--quiet"; ``True`` is "--progress"; ``False`` is "--progress" + "--quiet".
    only_ffmpeg : bool, default False
        Whether to use only FFmpeg (instead of yt-dlp) for URls.

    Returns
    -------
    numpy.ndarray
        A array containing the audio waveform in float32.
    """
    if sr is None:
        sr = SAMPLE_RATE
    file = load_source(file, verbose=verbose, only_ffmpeg=only_ffmpeg, return_dict=False)
    try:
        # This launches a subprocess to decode audio while down-mixing and resampling as necessary.
        # Requires the ffmpeg CLI in PATH.
        cmd = [
            "ffmpeg",
            "-nostdin",
            "-threads", "0",
            "-i", file if isinstance(file, str) else "pipe:",
            "-f", "s16le",
            "-ac", "1",
            "-acodec", "pcm_s16le",
            "-ar", str(sr),
            "-"
        ]
        if isinstance(file, str):
            out = subprocess.run(cmd, capture_output=True, check=True).stdout
        else:
            cmd = cmd[:1] + ["-loglevel", "error"] + cmd[1:]
            stdin = subprocess.PIPE if isinstance(file, bytes) else file
            out = subprocess.Popen(cmd, stdout=subprocess.PIPE, stdin=stdin)
            out = out.communicate(input=file if isinstance(file, bytes) else None)[0]
            if not out:
                raise RuntimeError(f"FFmpeg failed to load audio from bytes ({len(file)}).")
    except (subprocess.CalledProcessError, subprocess.SubprocessError) as e:
        raise RuntimeError(f"FFmpeg failed to load audio: {e.stderr.decode()}") from e

    return np.frombuffer(out, np.int16).flatten().astype(np.float32) / 32768.0


def resample(audio: torch.Tensor, in_sr: int, out_sr: int, **kwargs) -> torch.Tensor:
    return torchaudio.functional.resample(audio, in_sr, out_sr, **kwargs)


def voice_freq_filter(
        wf: Union[torch.Tensor, np.ndarray],
        sr: int,
        upper_freq: int = None,
        lower_freq: int = None
) -> torch.Tensor:
    if isinstance(wf, np.ndarray):
        wf = torch.from_numpy(wf)
    if upper_freq is None:
        upper_freq = 5000
    if lower_freq is None:
        lower_freq = 200
    assert upper_freq > lower_freq, f'upper_freq {upper_freq} must but greater than lower_freq {lower_freq}'
    return torchaudio.functional.highpass_biquad(
        torchaudio.functional.lowpass_biquad(wf, sr, upper_freq),
        sr,
        lower_freq
    )


def get_metadata(audiofile: Union[str, bytes, np.ndarray, torch.Tensor]) -> dict:
    if isinstance(audiofile, (np.ndarray, torch.Tensor)):
        return dict(sr=SAMPLE_RATE, duration=audiofile.shape[-1]/SAMPLE_RATE)
    import re
    cmd = ['ffmpeg', '-hide_banner', '-i']
    if isinstance(audiofile, str):
        cmd.append(audiofile)
        metadata = subprocess.run(
            cmd, capture_output=True,
        ).stderr.decode(errors="ignore")
    else:
        cmd.append('-')
        p = subprocess.Popen(
            cmd,  stderr=subprocess.PIPE, stdin=subprocess.PIPE,
        )
        try:
            p.stdin.write(audiofile)
        except BrokenPipeError:
            pass
        finally:
            metadata = p.communicate()[-1]
            if metadata is not None:
                metadata = metadata.decode(errors="ignore")
    sr = re.findall(r'\n.+Stream.+Audio.+\D+(\d+) Hz', metadata)
    duration = re.findall(r'Duration: ([\d:]+\.\d+),', metadata)
    if duration:
        h, m, s = duration[0].split(':')
        duration = int(h) * 3600 + int(m) * 60 + float(s)
    else:
        duration = None
    return dict(sr=int(sr[0]) if sr else None, duration=duration)


def get_samplerate(audiofile: Union[str, bytes]) -> Union[int, None]:
    return get_metadata(audiofile).get('sr')


def audio_to_tensor_resample(
        audio: Union[torch.Tensor, np.ndarray, str, bytes],
        original_sample_rate: Optional[int] = None,
        target_sample_rates: Optional[Union[int, Iterable[int]]] = None,
        **kwargs
) -> torch.Tensor:
    """
    Return ``audio`` as Tensor and resample as needed.
    """
    if target_sample_rates and isinstance(target_sample_rates, int):
        target_sample_rates = [target_sample_rates]

    if isinstance(audio, (str, bytes)):
        if target_sample_rates:
            original_sample_rate = target_sample_rates[0]
        audio = load_audio(audio, sr=original_sample_rate,**kwargs)
    elif not original_sample_rate:
        original_sample_rate = SAMPLE_RATE
    if isinstance(audio, np.ndarray):
        audio = torch.from_numpy(audio)
    audio = audio.float()

    if target_sample_rates and original_sample_rate not in target_sample_rates:
        audio = resample(audio, original_sample_rate, target_sample_rates[0])

    return audio


def standardize_audio(
        audio: Union[torch.Tensor, np.ndarray, str, bytes],
        resample_sr: Tuple[Optional[int], Union[int, Tuple[int]]] = None,
) -> torch.Tensor:
    warnings.warn('This `standardize_audio()` is deprecated and will be removed. '
                  'Use `stable_whisper.audio.utils.audio_to_tensor_resample()` instead.')
    return audio_to_tensor_resample(audio, *(resample_sr or ()))



import inspect
import sys

system_encoding = sys.getdefaultencoding()

if system_encoding != "utf-8":

    def make_safe(string):
        # replaces any character not representable using the system default encoding with an '?',
        # avoiding UnicodeEncodeError (https://github.com/openai/whisper/discussions/729).
        return string.encode(system_encoding, errors="replace").decode(system_encoding)

else:

    def make_safe(string):
        # utf-8 can encode any Unicode code point, so no need to do the round-trip encoding
        return string


def str_to_valid_type(val: str):
    if len(val) == 0:
        return None
    if '/' in val:
        return [a.split('*') if '*' in a else a for a in val.split('/')]
    try:
        val = float(val) if '.' in val else int(val)
    except ValueError:
        pass
    finally:
        return val


def get_func_parameters(func):
    return inspect.signature(func).parameters.keys()


def isolate_useful_options(options: dict, method, pop: bool = False) -> dict:
    _get = dict.pop if pop else dict.get
    return {k: _get(options, k) for k in get_func_parameters(method) if k in options}


def safe_print(msg: str, _print=None):
    if msg:
        (_print or print)(make_safe(msg))


def format_timestamp(
    seconds: float, always_include_hours: bool = False, decimal_marker: str = "."
):
    assert seconds >= 0, "non-negative timestamp expected"
    milliseconds = round(seconds * 1000.0)

    hours = milliseconds // 3_600_000
    milliseconds -= hours * 3_600_000

    minutes = milliseconds // 60_000
    milliseconds -= minutes * 60_000

    seconds = milliseconds // 1_000
    milliseconds -= seconds * 1_000

    hours_marker = f"{hours:02d}:" if always_include_hours or hours > 0 else ""
    return (
        f"{hours_marker}{minutes:02d}:{seconds:02d}{decimal_marker}{milliseconds:03d}"
    )


class UnsortedException(Exception):

    def __init__(self, message: str = None, data: dict = None):
        if not message:
            message = 'Timestamps are not in ascending order. If data is produced by Stable-ts, please submit an issue.'
        super().__init__(message)
        self.data = data

    def get_data(self):
        return self.data


def update_options(_options: dict = None, _reversed: bool = False, **kwargs) -> dict:
    if _options is None:
        _options = {}
    if _reversed:
        _options, kwargs = kwargs, _options
    _options.update(kwargs)
    return _options



import warnings
from typing import Optional, Union

import numpy as np
import torch

from ..audio import convert_demucs_kwargs, prep_audio
from ..non_whisper import transcribe_any
from ..utils import isolate_useful_options


HF_MODELS = {
    "tiny.en": "openai/whisper-tiny.en",
    "tiny": "openai/whisper-tiny",
    "base.en": "openai/whisper-base.en",
    "base": "openai/whisper-base",
    "small.en": "openai/whisper-small.en",
    "small": "openai/whisper-small",
    "medium.en": "openai/whisper-medium.en",
    "medium": "openai/whisper-medium",
    "large-v1": "openai/whisper-large-v1",
    "large-v2": "openai/whisper-large-v2",
    "large-v3": "openai/whisper-large-v3",
    "large": "openai/whisper-large-v3",
}


def get_device(device: str = None) -> str:
    if device:
        return device
    if torch.cuda.is_available():
        return 'cuda:0'
    if (mps := getattr(torch.backends, 'mps', None)) is not None and mps.is_available():
        return 'mps'
    return 'cpu'


def load_hf_pipe(model_name: str, device: str = None, flash: bool = False):
    from transformers import AutoModelForSpeechSeq2Seq, AutoProcessor, pipeline
    device = get_device(device)
    is_cpu = (device if isinstance(device, str) else getattr(device, 'type', None)) == 'cpu'
    dtype = torch.float32 if is_cpu or not torch.cuda.is_available() else torch.float16
    model_id = HF_MODELS.get(model_name, model_name)
    model = AutoModelForSpeechSeq2Seq.from_pretrained(
        model_id,
        torch_dtype=dtype,
        low_cpu_mem_usage=True,
        use_safetensors=True,
        use_flash_attention_2=flash
    ).to(device)

    processor = AutoProcessor.from_pretrained(model_id)

    if not flash:
        try:
            model = model.to_bettertransformer()
        except ValueError:
            pass

    pipe = pipeline(
        "automatic-speech-recognition",
        model=model,
        tokenizer=processor.tokenizer,
        feature_extractor=processor.feature_extractor,
        max_new_tokens=128,
        chunk_length_s=30,
        torch_dtype=dtype,
        device=device,
    )

    return pipe


class GenericWrapper:

    def __init__(self, obj, prompt_ids=None):
        self.obj = obj
        self.prompt_ids = None if prompt_ids is None or len(prompt_ids) == 0 else prompt_ids

    def __getattr__(self, attr):
        val = getattr(self.obj, attr)
        return val


class InnerIteratorWrapper(GenericWrapper):

    def __next__(self):
        output = next(self.obj)
        if self.prompt_ids is None:
            return output
        prompt_len = len(self.prompt_ids)
        for chunk in output:
            if (chunk['tokens'][..., :prompt_len] == self.prompt_ids).all():
                chunk['tokens'] = chunk['tokens'][..., prompt_len:]
            if 'token_timestamps' in chunk and (chunk['token_timestamps'][..., :prompt_len] == self.prompt_ids).all():
                chunk['token_timestamps'] = chunk['token_timestamps'][..., prompt_len:]
        return output


class IteratorWrapper(GenericWrapper):

    def __iter__(self):
        iter(self.obj)
        self.obj.iterator = InnerIteratorWrapper(self.obj.iterator, self.prompt_ids)
        return self.obj


class PipelineWrapper(GenericWrapper):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.super_mode = False
        self.call_kwargs = {}

    def __getattr__(self, attr):
        non_parent_val = super().__getattr__(attr)
        if not self.super_mode:
            return non_parent_val
        val = getattr(super(type(self.obj), self.obj), attr, non_parent_val)
        if hasattr(val, '__isabstractmethod__') and val.__isabstractmethod__:
            val = non_parent_val
        if attr == 'get_iterator':

            def get_iterator(*_args, **_kwargs):
                iterator = val(*_args, **_kwargs)
                iterator = IteratorWrapper(iterator, self.prompt_ids)
                return iterator

            return get_iterator

        elif attr == 'run_single':

            def run_single(*_args, **_kwargs):
                curr_class = self.obj.__class__
                if self.super_mode:
                    curr_class = curr_class.__base__
                if self.framework == 'pt' and curr_class.__name__ == 'ChunkPipeline':
                    _args = list(_args)
                    _args = (
                            [[_args[0]]] +
                            [
                                self.call_kwargs.get('num_workers') or self._num_workers or 0,
                                self.call_kwargs.get('batch_size') or self._batch_size or 1,
                            ] +
                            _args[1:]
                    )
                    output = next(iter(self.get_iterator(*_args, **_kwargs)))
                else:
                    output = val(*_args, **_kwargs)

                return output

            return run_single

        return val

    def __call__(self, *args, **kwargs):
        call = self.obj.__class__.__base__.__call__
        self.call_kwargs = kwargs
        self.super_mode = True
        return call(self, *args, **kwargs)


class WhisperHF:

    def __init__(self, model_name: str, device: str = None, flash: bool = False, pipeline=None):
        self._model_name = model_name
        self._pipe = load_hf_pipe(self._model_name, device, flash=flash) if pipeline is None else pipeline
        self._model_name = getattr(self._pipe.model, 'name_or_path', self._model_name)

    @property
    def sampling_rate(self):
        return self._pipe.feature_extractor.sampling_rate

    @property
    def model_name(self):
        if self._model_name is None:
            return getattr(self._pipe.model, 'name_or_path', 'n/a')
        return self._model_name

    def _modify_pipe_iterator(self, prompt_ids=None):
        if not isinstance(self._pipe, PipelineWrapper):
            self._pipe = PipelineWrapper(self._pipe, prompt_ids)

    def _inner_transcribe(
            self,
            audio: Union[str, bytes, np.ndarray],
            language: Optional[str] = None,
            task: Optional[str] = 'transcribe',
            batch_size: Optional[int] = 24,
            word_timestamps=True,
            verbose: Optional[bool] = False,
            initial_prompt: Optional[str] = None,
            **kwargs
    ):
        generate_kwargs = {}
        if self.model_name.endswith('en'):
            language = task = None
        if task is not None:
            generate_kwargs['task'] = task
        if language is not None:
            generate_kwargs['language'] = language
        prompt_ids = kwargs.pop('prompt_ids', None)
        if initial_prompt:
            if prompt_ids:
                warnings.warn(
                    'Got both ``prompt_ids`` and ``initial_prompt`` but ``prompt_ids`` will be ignored.',
                    stacklevel=2
                )
            prompt_ids = self._pipe.tokenizer.get_prompt_ids(initial_prompt, 'pt')
        if prompt_ids is not None:
            self._modify_pipe_iterator(prompt_ids)
            generate_kwargs['prompt_ids'] = prompt_ids
        generate_kwargs.update(kwargs)
        if verbose is not None:
            print(f'Transcribing with Hugging Face Whisper ({self.model_name})...')
        pipe_kwargs = dict(
            generate_kwargs=generate_kwargs,
            return_timestamps='word' if word_timestamps else True
        )
        if batch_size is not None:
            pipe_kwargs['batch_size'] = batch_size
        output = self._pipe(audio, **pipe_kwargs)
        result = output['chunks']
        if verbose is not None:
            print(f'Transcription completed.')

        def replace_none_ts(parts):
            total_dur = round(audio.shape[-1] / self.sampling_rate, 3) if isinstance(audio, np.ndarray) else None
            _medium_dur = _ts_nonzero_mask = None

            def ts_nonzero_mask() -> np.ndarray:
                nonlocal _ts_nonzero_mask
                if _ts_nonzero_mask is None:
                    _ts_nonzero_mask = np.array([(p['end'] or p['start']) is not None for p in parts])
                return _ts_nonzero_mask

            def medium_dur() -> float:
                nonlocal _medium_dur
                if _medium_dur is None:
                    nonzero_dus = [p['end'] - p['start'] for p in parts if None not in (p['end'], p['start'])]
                    nonzero_durs = np.array(nonzero_dus)
                    _medium_dur = np.median(nonzero_durs) * 2 if len(nonzero_durs) else 2.0
                return _medium_dur

            def _curr_max_end(start: float, next_idx: float) -> float:
                max_end = total_dur
                if next_idx != len(parts):
                    mask = np.flatnonzero(ts_nonzero_mask()[next_idx:])
                    if len(mask):
                        _part = parts[mask[0]+next_idx]
                        max_end = _part['start'] or _part['end']

                new_end = round(start + medium_dur(), 3)
                if max_end is None:
                    return new_end
                if new_end > max_end:
                    return max_end
                return new_end

            for i, part in enumerate(parts, 1):
                if part['start'] is None:
                    is_first = i == 1
                    if is_first:
                        new_start = round((part['end'] or 0) - medium_dur(), 3)
                        part['start'] = max(new_start, 0.0)
                    else:
                        part['start'] = parts[i - 2]['end']
                if part['end'] is None:
                    no_next_start = i == len(parts) or parts[i]['start'] is None
                    part['end'] = _curr_max_end(part['start'], i) if no_next_start else parts[i]['start']

        if word_timestamps:
            words = [
                dict(start=word['timestamp'][0], end=word['timestamp'][1], word=word['text'])
                for word in result
            ]
            replace_none_ts(words)
            return [words]
        segs = [
            dict(start=seg['timestamp'][0], end=seg['timestamp'][1], text=seg['text'])
            for seg in result
        ]
        replace_none_ts(segs)
        return segs

    def transcribe(
            self,
            audio: Union[str, bytes, np.ndarray],
            *,
            language: Optional[str] = None,
            task: Optional[str] = 'transcribe',
            batch_size: Optional[int] = 24,
            word_timestamps: bool = True,
            verbose: Optional[bool] = False,
            regroup: Union[bool, str] = True,
            suppress_silence: bool = True,
            suppress_word_ts: bool = True,
            use_word_position: bool = True,
            q_levels: int = 20,
            k_size: int = 5,
            denoiser: Optional[str] = None,
            denoiser_options: Optional[dict] = None,
            vad: bool = False,
            vad_threshold: float = 0.35,
            vad_onnx: bool = False,
            min_word_dur: Optional[float] = None,
            nonspeech_error: float = 0.1,
            only_voice_freq: bool = False,
            only_ffmpeg: bool = False,
            check_sorted: bool = True,
            **options
    ):
        transcribe_any_options = isolate_useful_options(options, transcribe_any, pop=True)
        denoiser, denoiser_options = convert_demucs_kwargs(
            denoiser, denoiser_options,
            demucs=transcribe_any_options.pop('demucs', None),
            demucs_options=transcribe_any_options.pop('demucs_options', None)
        )

        if isinstance(audio, (str, bytes)):
            audio = prep_audio(audio, sr=self.sampling_rate).numpy()
            transcribe_any_options['input_sr'] = self.sampling_rate

        if 'input_sr' not in transcribe_any_options:
            transcribe_any_options['input_sr'] = self.sampling_rate

        if denoiser or only_voice_freq:
            if 'audio_type' not in transcribe_any_options:
                transcribe_any_options['audio_type'] = 'numpy'
            if 'model_sr' not in transcribe_any_options:
                transcribe_any_options['model_sr'] = self.sampling_rate

        inference_kwargs = dict(
            audio=audio,
            language=language,
            task=task,
            batch_size=batch_size,
            word_timestamps=word_timestamps,
            verbose=verbose,
            **options
        )
        return transcribe_any(
            inference_func=self._inner_transcribe,
            audio=audio,
            inference_kwargs=inference_kwargs,
            verbose=verbose,
            regroup=regroup,
            suppress_silence=suppress_silence,
            suppress_word_ts=suppress_word_ts,
            q_levels=q_levels,
            k_size=k_size,
            denoiser=denoiser,
            denoiser_options=denoiser_options,
            vad=vad,
            vad_threshold=vad_threshold,
            vad_onnx=vad_onnx,
            min_word_dur=min_word_dur,
            nonspeech_error=nonspeech_error,
            use_word_position=use_word_position,
            only_voice_freq=only_voice_freq,
            only_ffmpeg=only_ffmpeg,
            force_order=True,
            check_sorted=check_sorted,
            **transcribe_any_options
        )


def load_hf_whisper(model_name: str, device: str = None, flash: bool = False, pipeline=None):
    return WhisperHF(model_name, device, flash=flash, pipeline=pipeline)



import string
import torch
import numpy as np
from typing import TYPE_CHECKING, List, Callable, Optional
from itertools import chain

from whisper.audio import TOKENS_PER_SECOND, N_SAMPLES_PER_TOKEN
from whisper.timing import WordTiming, median_filter, dtw, merge_punctuations

if TYPE_CHECKING:
    from whisper.tokenizer import Tokenizer
    from whisper.model import Whisper


# modified version of whisper.timing.find_alignment
def find_alignment_stable(
        model: "Whisper",
        tokenizer: "Tokenizer",
        text_tokens: List[int],
        mel: torch.Tensor,
        num_samples: int,
        *,
        medfilt_width: int = 7,
        qk_scale: float = 1.0,
        ts_num: int = 0,
        ts_noise: float = 0.1,
        token_split=None,
        audio_features: torch.Tensor = None,
        ts_token_mask: torch.Tensor = None
) -> List[WordTiming]:
    tokens = torch.tensor(
        [
            *tokenizer.sot_sequence,
            tokenizer.no_timestamps,
            *text_tokens,
            tokenizer.eot,
        ]
    ).to(model.device)

    # install hooks on the cross attention layers to retrieve the attention weights
    QKs = [None] * model.dims.n_text_layer
    hooks = [
        block.cross_attn.register_forward_hook(
            lambda _, ins, outs, index=i: QKs.__setitem__(index, outs[-1])
        )
        for i, block in enumerate(model.decoder.blocks)
    ]

    with torch.no_grad():
        if audio_features is None:
            audio_features = model.encoder(mel.unsqueeze(0))
        if ts_num:
            if ts_noise is None:
                ts_noise = 0.1
            extra_audio_features = audio_features.repeat_interleave(ts_num, 0)
            torch.manual_seed(0)
            audio_features = torch.cat([audio_features,
                                        extra_audio_features *
                                        (1 - (torch.rand_like(extra_audio_features) * ts_noise))],
                                       dim=0)
            logits = model.decoder(tokens.unsqueeze(0).repeat_interleave(audio_features.shape[0], 0),
                                   audio_features)
        else:
            logits = model.decoder(tokens.unsqueeze(0), audio_features)

        logits = logits[0]
        sampled_logits = logits[len(tokenizer.sot_sequence):, : tokenizer.eot]
        token_probs = sampled_logits.softmax(dim=-1)
        text_token_probs = token_probs[np.arange(len(text_tokens)), text_tokens]
        text_token_probs = text_token_probs.tolist()

    for hook in hooks:
        hook.remove()

    # heads * tokens * frames
    weights = torch.cat([QKs[_l][:, _h] for _l, _h in model.alignment_heads.indices().T], dim=0)
    weights = weights[:, :, : round(num_samples / N_SAMPLES_PER_TOKEN)]
    weights = (weights * qk_scale).softmax(dim=-1)
    std, mean = torch.std_mean(weights, dim=-2, keepdim=True, unbiased=False)
    weights = (weights - mean) / std
    weights = median_filter(weights, medfilt_width)

    matrix = weights.mean(axis=0)
    matrix = matrix[len(tokenizer.sot_sequence): -1]
    if ts_token_mask is not None:
        matrix[..., ts_token_mask[:matrix.shape[-1]]] = 0
    text_indices, time_indices = dtw(-matrix)

    if token_split is None:
        words, word_tokens = tokenizer.split_to_word_tokens(text_tokens + [tokenizer.eot])
    else:
        words, word_tokens = token_split
        words.append(tokenizer.decode([tokenizer.eot]))
        word_tokens.append([tokenizer.eot])
    word_boundaries = np.pad(np.cumsum([len(t) for t in word_tokens[:-1]]), (1, 0))

    jumps = np.pad(np.diff(text_indices), (1, 0), constant_values=1).astype(bool)
    jump_times = time_indices[jumps].clip(min=0) / TOKENS_PER_SECOND
    start_times = jump_times[word_boundaries[:-1]]
    end_times = jump_times[word_boundaries[1:]]
    word_probabilities = [
        np.mean(text_token_probs[i:j])
        for i, j in zip(word_boundaries[:-1], word_boundaries[1:])
    ]

    return [
        WordTiming(word, tokens, start, end, probability)
        for word, tokens, start, end, probability in zip(
            words, word_tokens, start_times, end_times, word_probabilities
        )
    ]


def _split_tokens(tokens: List[int], tokenizer: "Tokenizer"):
    split_by_space = getattr(tokenizer, 'language_code', tokenizer.language) not in {"zh", "ja", "th", "lo", "my"}
    text = tokenizer.decode_with_timestamps(tokens)
    words = []
    word_tokens = []
    curr_tokens = []
    is_append = False
    for token in tokens:
        curr_tokens.append(token)
        curr_text = tokenizer.decode(curr_tokens)
        is_whole = token >= tokenizer.eot
        if not is_whole:
            is_whole = text[:len(curr_text)] == curr_text
            if is_whole and split_by_space:
                is_append = not (curr_text.startswith(" ") or curr_text.strip() in string.punctuation)

        if is_whole:
            if is_append and len(words) != 0:
                words[-1] += curr_text
                word_tokens[-1].extend(curr_tokens)
            else:
                words.append(curr_text)
                word_tokens.append(curr_tokens)
            text = text[len(curr_text):]
            curr_tokens = []

    if len(curr_tokens) != 0:
        words.append(curr_text if len(text) == 0 else text)
        word_tokens.append(curr_tokens)
    elif len(text) != 0:
        words[-1] += text

    return words, word_tokens


def split_word_tokens(segments: List[dict],
                      tokenizer: "Tokenizer",
                      *,
                      padding: (str, int) = None,
                      split_callback: Callable = None):
    if padding is not None:
        if isinstance(padding, str):
            padding = tokenizer.encode(padding)
        else:
            padding = [padding]
    tokens = []
    seg_indices = []
    words = []
    word_tokens = []
    for i, s in enumerate(segments):
        temp_word_tokens = [t for t in s['tokens'] if not isinstance(t, int) or t < tokenizer.eot]
        curr_words, curr_word_tokens = (
            _split_tokens(temp_word_tokens, tokenizer)
            if split_callback is None else
            split_callback(temp_word_tokens, tokenizer)
        )
        assert len(curr_words) == len(curr_word_tokens), \
            f'word count and token group count do not match, {len(curr_words)} and {len(curr_word_tokens)}'
        if (
                padding is not None and
                curr_word_tokens[0][0] != padding and
                (len(tokens) == 0 or tokens[-1] != padding)
        ):
            tokens.extend(padding)
            words.append(None)
            word_tokens.append(padding)
        seg_indices.extend([i] * len(curr_words))
        tokens.extend(list(chain.from_iterable(curr_word_tokens)))
        words.extend(curr_words)
        word_tokens.extend(curr_word_tokens)

    return tokens, (words, word_tokens), seg_indices


def pop_empty_alignment(alignment: List[WordTiming]):
    return list(reversed([alignment.pop(i) for i in reversed(range(len(alignment))) if alignment[i].word is None]))


# modified version of whisper.timing.add_word_timestamps
def add_word_timestamps_stable(
        *,
        segments: List[dict],
        model: "Whisper",
        tokenizer: "Tokenizer",
        mel: torch.Tensor,
        num_samples: int,
        prepend_punctuations: str = "\"'“¿([{-",
        append_punctuations: str = "\"'.。,，!！?？:：”)]}、",
        audio_features: torch.Tensor = None,
        ts_num: int = 0,
        ts_noise: float = 0.1,
        min_word_dur: float = 0.1,
        split_callback: Callable = None,
        gap_padding: Optional[str] = ' ...',
        ts_token_mask: torch.Tensor = None,
        **kwargs,
):
    if len(segments) == 0:
        return

    if min_word_dur is None:
        min_word_dur = 0

    if prepend_punctuations is None:
        prepend_punctuations = "\"'“¿([{-"

    if append_punctuations is None:
        append_punctuations = "\"'.。,，!！?？:：”)]}、"

    def align():
        for seg in segments:
            seg['words'] = []

        text_tokens, token_split, seg_indices = split_word_tokens(segments, tokenizer,
                                                                  padding=gap_padding, split_callback=split_callback)

        alignment = find_alignment_stable(model, tokenizer, text_tokens, mel, num_samples,
                                          **kwargs,
                                          token_split=token_split,
                                          audio_features=audio_features,
                                          ts_num=ts_num,
                                          ts_noise=ts_noise,
                                          ts_token_mask=ts_token_mask)
        alt_beginning_alignment = pop_empty_alignment(alignment)

        merge_punctuations(alignment, prepend_punctuations, append_punctuations)

        time_offset = segments[0]["seek"]

        assert len(alignment) == len(seg_indices)
        assert (gap_padding is None or len(segments) == len(alt_beginning_alignment))
        for i, timing in zip(seg_indices, alignment):
            if len(timing.tokens) != 0:
                start = timing.start
                end = timing.end
                if (
                        len(segments[i]['words']) == 0 and
                        ((end - start) < min_word_dur) and
                        len(alt_beginning_alignment)
                ):
                    start = alt_beginning_alignment[i].start
                segments[i]['words'].append(
                    dict(
                        word=timing.word,
                        start=round(time_offset + start, 3),
                        end=round(time_offset + end, 3),
                        probability=timing.probability,
                        tokens=timing.tokens
                    )
                )

    align()
    if (
            gap_padding is not None and
            any(
                (word['end'] - word['start']) < min_word_dur
                for seg in segments
                for word in seg['words']
            )
    ):
        gap_padding = None
        align()

    for segment in segments:
        if len(words := segment["words"]) > 0:
            # adjust the segment-level timestamps based on the word-level timestamps
            segment["start"] = words[0]["start"]
            segment["end"] = words[-1]["end"]



import warnings
import importlib.metadata

import whisper.tokenizer

from .utils import get_func_parameters

_COMPATIBLE_WHISPER_VERSIONS = (
    '20230314',
    '20230918',
    '20231105',
    '20231106',
    '20231117',
)
_required_whisper_ver = _COMPATIBLE_WHISPER_VERSIONS[-1]

_TOKENIZER_PARAMS = get_func_parameters(whisper.tokenizer.get_tokenizer)


def warn_compatibility_issues(
        whisper_module,
        ignore: bool = False,
        additional_msg: str = ''
):
    compatibility_warning = ''
    if not ignore:
        if whisper_module.__version__ not in _COMPATIBLE_WHISPER_VERSIONS:
            compatibility_warning += (f'Whisper {whisper_module.__version__} is installed.'
                                      f'Versions confirm to be compatible: {", ".join(_COMPATIBLE_WHISPER_VERSIONS)}\n')
        _is_whisper_repo_version = bool(importlib.metadata.distribution('openai-whisper').read_text('direct_url.json'))
        if _is_whisper_repo_version:
            compatibility_warning += ('The detected version appears to be installed from the repository '
                                      'which can have compatibility issues '
                                      'due to multiple commits sharing the same version number. '
                                      f'It is recommended to install version {_required_whisper_ver} from PyPI.\n')

        if compatibility_warning:
            compatibility_warning = (
                    'The installed version of Whisper might be incompatible.\n'
                    + compatibility_warning +
                    'To prevent errors and performance issues, reinstall correct version with: '
                    f'"pip install --upgrade --no-deps --force-reinstall openai-whisper=={_required_whisper_ver}".'
            )
            if additional_msg:
                compatibility_warning += f' {additional_msg}'
            warnings.warn(compatibility_warning)


def get_tokenizer(model=None, is_faster_model: bool = False, **kwargs):
    """
    Backward compatible wrapper of :func:`whisper.tokenizer.get_tokenizer` and
    :class:`faster_whisper.tokenizer.Tokenizer`.
    """
    if is_faster_model:
        import faster_whisper.tokenizer
        tokenizer = faster_whisper.tokenizer.Tokenizer
        params = get_func_parameters(tokenizer)
        if model is not None and 'tokenizer' not in kwargs:
            kwargs['tokenizer'] = model.hf_tokenizer
    else:
        tokenizer = whisper.tokenizer.get_tokenizer
        params = _TOKENIZER_PARAMS
    if model is not None and 'multilingual' not in kwargs:
        kwargs['multilingual'] = \
            (model.is_multilingual if hasattr(model, 'is_multilingual') else model.model.is_multilingual)
    if 'num_languages' in params:
        if hasattr(model, 'num_languages'):
            kwargs['num_languages'] = \
                (model.num_languages if hasattr(model, 'num_languages') else model.model.num_languages)
    elif 'num_languages' in kwargs:
        del kwargs['num_languages']
    return tokenizer(**kwargs)



import warnings
from itertools import chain
from typing import List, Union, Optional, Tuple

import numpy as np
import torch

from whisper.audio import TOKENS_PER_SECOND


def is_ascending_sequence(
        seq: List[Union[int, float]],
        verbose=True
) -> bool:
    """
    Return boolean for whether a sequence of numbers is in ascending order.
    """
    is_ascending = True
    for idx, (i, j) in enumerate(zip(seq[:-1], seq[1:])):
        if i > j:
            is_ascending = False
            if verbose:
                print(f'[Index{idx}]:{i} > [Index{idx + 1}]:{j}')
            else:
                break

    return is_ascending


def valid_ts(
        ts: List[dict],
        warn=True
) -> bool:
    """
    Return boolean for whether a list of timestamps is in ascending order.
    """
    valid = is_ascending_sequence(list(chain.from_iterable([s['start'], s['end']] for s in ts)), False)
    if warn and not valid:
        warnings.warn(message='Found timestamp(s) out of order.')
    return valid


def mask2timing(
        silence_mask: (np.ndarray, torch.Tensor),
        time_offset: float = 0.0,
        second_per_unit: Optional[float] = None,
        min_start: Optional[float] = None,
        max_end: Optional[float] = None
) -> (Tuple[np.ndarray, np.ndarray], None):
    """
    Return array of start timestamps and array of end timestamps corresponding to where ``silence_mask`` is ``True``.
    """
    if silence_mask is None or not silence_mask.any() or not len(silence_mask):
        return
    assert silence_mask.ndim == 1
    if isinstance(silence_mask, torch.Tensor):
        silence_mask = silence_mask.cpu().numpy().copy()
    elif not isinstance(silence_mask, np.ndarray):
        raise NotImplementedError(f'Expected torch.Tensor or numpy.ndarray, but got {type(silence_mask)}')

    mask = np.concatenate(([False], silence_mask, [False]))
    silent_starts = np.logical_and(~mask[:-2], mask[1:-1]).nonzero()[0]
    silent_ends = (np.logical_and(mask[1:-1], ~mask[2:]).nonzero()[0] + 1)
    if second_per_unit is None:
        silent_starts = silent_starts / TOKENS_PER_SECOND
        silent_ends = silent_ends / TOKENS_PER_SECOND
    else:
        silent_starts = silent_starts * second_per_unit
        silent_ends = silent_ends * second_per_unit
    if time_offset:
        silent_starts += time_offset
        silent_ends += time_offset
    if min_start is not None and silent_starts[0] < min_start:
        assert min_start <= silent_ends[0]
        silent_starts[0] = min_start
    if max_end is not None and silent_ends[-1] > max_end:
        assert max_end >= silent_starts[-1]
        silent_ends[-1] = max_end
    return silent_starts, silent_ends


def timing2mask(
        silent_starts: np.ndarray,
        silent_ends: np.ndarray,
        size: int,
        time_offset: float = None,
        units_per_second: Optional[int] = None
) -> torch.Tensor:
    """
    Return Tensor of booleans that is ``True`` corresponding to ``silent_starts`` and ``silent_ends``.
    """
    if units_per_second is None:
        units_per_second = TOKENS_PER_SECOND
    assert len(silent_starts) == len(silent_ends)
    ts_token_mask = torch.zeros(size, dtype=torch.bool)
    if time_offset:
        silent_starts = (silent_starts - time_offset).clip(min=0)
        silent_ends = (silent_ends - time_offset).clip(min=0)
    mask_i = (silent_starts * units_per_second).round().astype(np.int32)
    mask_e = (silent_ends * units_per_second).round().astype(np.int32)
    for mi, me in zip(mask_i, mask_e):
        ts_token_mask[mi:me+1] = True

    return ts_token_mask


class SetTorchThread:
    def __init__(self, temp_thread_count: int):
        self.original_thread_count = torch.get_num_threads()
        self.temp_thread_count = temp_thread_count

    def __enter__(self):
        torch.set_num_threads(self.temp_thread_count)

    def __exit__(self, exc_type, exc_val, exc_tb):
        torch.set_num_threads(self.original_thread_count)



import torch
from torch import nn
from whisper.model import Linear, Conv1d, LayerNorm, Whisper


def replace_modules(model: nn.Module, only_linear: bool = False):
    """
    Replace ``Linear``/``Conv1d``/``LayerNorm`` from :class:`whisper.model` with equivalent module in
        :class:`torch.nn`.
    """
    for m in model.__dict__.get('_modules', []):
        module = model.__getattr__(m)
        update = True
        if isinstance(module, Linear):
            model.__setattr__(m, nn.Linear(module.in_features, module.out_features,
                                           bias=module.bias is not None))
        elif not only_linear and isinstance(module, Conv1d):
            model.__setattr__(m, nn.Conv1d(module.in_channels, module.out_channels,
                                           kernel_size=module.kernel_size,
                                           stride=module.stride,
                                           padding=module.padding,
                                           bias=module.bias is not None))
        elif not only_linear and isinstance(module, LayerNorm):
            model.__setattr__(m, nn.LayerNorm(module.normalized_shape[0]))
        else:
            update = False
            replace_modules(module)

        if update:
            model.__getattr__(m).load_state_dict(module.state_dict())


def ptdq_linear(model: "Whisper"):
    """
    Apply Dynamic Quantization to instance of :class:`whisper.model.Whisper`.
    """
    model.cpu()
    replace_modules(model, only_linear=True)
    torch.quantization.quantize_dynamic(model, {nn.Linear}, dtype=torch.qint8, inplace=True)
    setattr(model, 'dq', True)



import os
import warnings
import io
import torch
import torchaudio
import numpy as np
from typing import Union, Callable, Optional

from .audio import AudioLoader, get_denoiser_func, convert_demucs_kwargs
from .audio.utils import load_source, load_audio, voice_freq_filter, get_samplerate, resample, audio_to_tensor_resample
from .result import WhisperResult
from .utils import update_options

AUDIO_TYPES = ('str', 'byte', 'torch', 'numpy')

AUDIO_TYPE_BY_CLASS = {
    str: 'str',
    bytes: 'bytes',
    np.ndarray: 'numpy',
    torch.Tensor: 'pytorch',
    AudioLoader: None
}


def transcribe_any(
        inference_func: Callable,
        audio: Union[str, np.ndarray, torch.Tensor, bytes, AudioLoader],
        audio_type: str = None,
        input_sr: int = None,
        model_sr: int = None,
        inference_kwargs: dict = None,
        temp_file: str = None,
        verbose: Optional[bool] = False,
        regroup: Union[bool, str] = True,
        suppress_silence: bool = True,
        suppress_word_ts: bool = True,
        q_levels: int = 20,
        k_size: int = 5,
        denoiser: Optional[str] = None,
        denoiser_options: Optional[dict] = None,
        demucs: bool = False,
        demucs_options: dict = None,
        vad: bool = False,
        vad_threshold: float = 0.35,
        vad_onnx: bool = False,
        min_word_dur: Optional[float] = None,
        nonspeech_error: float = 0.1,
        use_word_position: bool = True,
        only_voice_freq: bool = False,
        only_ffmpeg: bool = False,
        force_order: bool = False,
        check_sorted: bool = True
) -> WhisperResult:
    """
    Transcribe ``audio`` using any ASR system.

    Parameters
    ----------
    inference_func : Callable
        Function that runs ASR when provided the ``audio`` and return data in the appropriate format.
        For format examples see, https://github.com/jianfch/stable-ts/blob/main/examples/non-whisper.ipynb.
    audio : str or numpy.ndarray or torch.Tensor or bytes or AudioLoader
        Path/URL to the audio file, the audio waveform, bytes of audio file or
        instance of :class:`stable_whisper.audio.AudioLoader`.
    audio_type : {'str', 'byte', 'torch', 'numpy', None}, default None, meaning same type as ``audio``
        The type that ``audio`` needs to be for ``inference_func``.
        'str' is a path to the file.
        'byte' is bytes (used for APIs or to avoid writing any data to hard drive).
        'torch' is an instance of :class:`torch.Tensor` containing the audio waveform, in float32 dtype, on CPU.
        'numpy' is an instance of :class:`numpy.ndarray` containing the audio waveform, in float32 dtype.
    input_sr : int, default None, meaning auto-detected if ``audio`` is ``str`` or ``bytes``
        The sample rate of ``audio``.
    model_sr : int, default None, meaning same sample rate as ``input_sr``
        The sample rate to resample the audio into for ``inference_func``.
    inference_kwargs : dict, optional
        Dictionary of arguments to pass into ``inference_func``.
    temp_file : str, default './_temp_stable-ts_audio_.wav'
        Temporary path for the preprocessed audio when ``audio_type = 'str'``.
    verbose: bool, False
        Whether to displays all the details during transcription, If ``False``, displays progressbar. If ``None``, does
        not display anything.
    regroup: str or bool, default True
         String representation of a custom regrouping algorithm or ``True`` use to the default algorithm 'da'. Only
         applies if ``word_timestamps = False``.
    suppress_silence : bool, default True
        Whether to enable timestamps adjustments based on the detected silence.
    suppress_word_ts : bool, default True
        Whether to adjust word timestamps based on the detected silence. Only enabled if ``suppress_silence = True``.
    q_levels : int, default 20
        Quantization levels for generating timestamp suppression mask; ignored if ``vad = true``.
        Acts as a threshold to marking sound as silent.
        Fewer levels will increase the threshold of volume at which to mark a sound as silent.
    k_size : int, default 5
        Kernel size for avg-pooling waveform to generate timestamp suppression mask; ignored if ``vad = true``.
        Recommend 5 or 3; higher sizes will reduce detection of silence.
    denoiser : str, optional
        String of the denoiser to use for preprocessing ``audio``.
        See ``stable_whisper.audio.SUPPORTED_DENOISERS`` for supported denoisers.
    denoiser_options : dict, optional
        Options to use for ``denoiser``.
    vad : bool, default False
        Whether to use Silero VAD to generate timestamp suppression mask.
        Silero VAD requires PyTorch 1.12.0+. Official repo, https://github.com/snakers4/silero-vad.
    vad_threshold : float, default 0.35
        Threshold for detecting speech with Silero VAD. Low threshold reduces false positives for silence detection.
    vad_onnx : bool, default False
        Whether to use ONNX for Silero VAD.
    min_word_dur : float or None, default None meaning use ``stable_whisper.default.DEFAULT_VALUES``
        Shortest duration each word is allowed to reach for silence suppression.
    nonspeech_error : float, default 0.1
        Relative error of non-speech sections that appear in between a word for silence suppression.
    use_word_position : bool, default True
        Whether to use position of the word in its segment to determine whether to keep end or start timestamps if
        adjustments are required. If it is the first word, keep end. Else if it is the last word, keep the start.
    only_voice_freq : bool, default False
        Whether to only use sound between 200 - 5000 Hz, where majority of human speech are.
    only_ffmpeg : bool, default False
        Whether to use only FFmpeg (instead of not yt-dlp) for URls
    force_order : bool, default False
        Whether to use adjacent timestamps to replace timestamps that are out of order. Use this parameter only if
        the words/segments returned by ``inference_func`` are expected to be in chronological order.
    check_sorted : bool, default True
        Whether to raise an error when timestamps returned by ``inference_func`` are not in ascending order.

    Returns
    -------
    stable_whisper.result.WhisperResult
        All timestamps, words, probabilities, and other data from the transcription of ``audio``.

    Notes
    -----
    For ``audio_type = 'str'``:
        If ``audio`` is a file and no audio preprocessing is set, ``audio`` will be directly passed into
            ``inference_func``.
        If audio preprocessing is ``denoiser`` or ``only_voice_freq``, the processed audio will be encoded into
            ``temp_file`` and then passed into ``inference_func``.

    For ``audio_type = 'byte'``:
        If ``audio`` is file, the bytes of file will be passed into ``inference_func``.
        If ``audio`` is :class:`torch.Tensor` or :class:`numpy.ndarray`, the bytes of the ``audio`` will be encoded
            into WAV format then passed into ``inference_func``.

    Resampling is only performed on ``audio`` when ``model_sr`` does not match the sample rate of the ``audio`` before
        passing into ``inference_func`` due to ``input_sr`` not matching ``model_sr``, or sample rate changes due to
        audio preprocessing from ``denoiser``.
    """
    denoiser, denoiser_options = convert_demucs_kwargs(
        denoiser, denoiser_options, demucs=demucs, demucs_options=demucs_options
    )
    if audio_type is not None and (audio_type := audio_type.lower()) not in AUDIO_TYPES:
        raise NotImplementedError(f'``audio_type="{audio_type}"`` is not supported. Types: {AUDIO_TYPES}')

    if isinstance(audio, AudioLoader) and audio_type is not None:
        raise ValueError(f'``audio_type`` can only be ``None`` when ``audio`` is an AudioLoader instance,'
                         f'but got {audio_type}')

    if audio_type is None:
        if type(audio) in AUDIO_TYPE_BY_CLASS:
            audio_type = AUDIO_TYPE_BY_CLASS[type(audio)]
        else:
            raise TypeError(f'{type(audio)} is not supported for ``audio``.')

    if (
            input_sr is None and
            isinstance(audio, (np.ndarray, torch.Tensor)) and
            (denoiser or only_voice_freq or suppress_silence or model_sr)
    ):
        raise ValueError('``input_sr`` is required when ``audio`` is a PyTorch tensor or NumPy array.')

    if (
            model_sr is None and
            isinstance(audio, (str, bytes)) and
            audio_type in ('torch', 'numpy')
    ):
        raise ValueError('``model_sr`` is required when ``audio_type`` is a "pytorch" or "numpy".')

    if isinstance(audio, str):
        audio = load_source(audio, verbose=verbose, only_ffmpeg=only_ffmpeg)

    if inference_kwargs is None:
        inference_kwargs = {}

    temp_file = os.path.abspath(temp_file or './_temp_stable-ts_audio_.wav')
    temp_audio_file = None

    if isinstance(audio, AudioLoader):
        if denoiser and not audio._denoiser:
            warnings.warn(
                '``denoiser`` will have no affect unless specified at AudioLoader initialization.',
                stacklevel=2
            )
        denoiser = None

        if only_voice_freq and not audio._only_voice_freq:
            warnings.warn(
                '``only_voice_freq=True`` will have no affect unless specified at AudioLoader initialization.',
                stacklevel=2
            )
        only_voice_freq = False

        if suppress_silence:
            warnings.warn(
                '``suppress_silence=True`` is not yet supported when ``audio`` is an AudioLoader.',
                stacklevel=2
            )
        suppress_silence = False

        if input_sr is not None and input_sr != audio.sr:
            warnings.warn(
                f'``input_sr`` ({input_sr}) does not match ``sr`` of AudioLoader ({audio.sr})',
                stacklevel=2
            )
        input_sr = audio.sr

    is_audio_encoded = isinstance(audio, (str, bytes))

    audio_sr = input_sr

    def curr_audio_sr(is_optional: bool = False):
        nonlocal audio_sr
        if is_optional and is_audio_encoded:
            return None
        if audio_sr is not None:
            return audio_sr
        assert isinstance(audio, (str, bytes)), f'No ``input_sr`` specified.'
        audio_sr = get_samplerate(audio)
        assert audio_sr is not None, 'Failed to get samplerate from ``audio``'
        return audio_sr

    if denoiser:
        denoise_model = denoiser_options.pop('model', None)
        if denoise_model is None:
            denoise_model = get_denoiser_func(denoiser, 'load')(True)
    else:
        denoise_model = None

    if denoiser:
        denoiser_options = update_options(
            denoiser_options,
            True,
            audio=torch.from_numpy(audio) if isinstance(audio, np.ndarray) else audio,
            input_sr=curr_audio_sr(True),
            model=denoise_model,
            verbose=verbose
        )
        denoise_run = get_denoiser_func(denoiser, 'run')
        audio = denoise_run(**denoiser_options)
        audio_sr = denoise_model.samplerate
        is_audio_encoded = False
        if (denoise_output := denoiser_options.get('save_path')) and audio_type == 'str':
            audio = denoise_output

    if only_voice_freq:
        if is_audio_encoded and audio_sr and model_sr:
            audio_sr = max(audio_sr, model_sr)
        audio = audio_to_tensor_resample(
            audio,
            original_sample_rate=curr_audio_sr(),
            verbose=verbose,
            only_ffmpeg=only_ffmpeg
        )
        audio = voice_freq_filter(audio, audio_sr)
        is_audio_encoded = False

    final_audio = audio

    if model_sr is None:
        final_audio_sr = audio_sr
    else:
        final_audio_sr = curr_audio_sr()

        if final_audio_sr != model_sr:
            if isinstance(final_audio, (str, bytes)):
                final_audio = load_audio(
                    final_audio,
                    sr=model_sr,
                    verbose=verbose,
                    only_ffmpeg=only_ffmpeg
                )
                final_audio_sr = model_sr
            else:
                if isinstance(final_audio, np.ndarray):
                    final_audio = torch.from_numpy(final_audio)
                if isinstance(final_audio, torch.Tensor):
                    final_audio = resample(final_audio, audio_sr, model_sr)
                    final_audio_sr = model_sr

    if audio_type in ('torch', 'numpy'):

        if isinstance(final_audio, (str, bytes)):
            final_audio = load_audio(
                final_audio,
                sr=model_sr,
                verbose=verbose,
                only_ffmpeg=only_ffmpeg
            )

        if not isinstance(final_audio, AudioLoader):
            if audio_type == 'torch':
                if isinstance(final_audio, np.ndarray):
                    final_audio = torch.from_numpy(final_audio)
            elif isinstance(final_audio, torch.Tensor):
                final_audio = final_audio.cpu().numpy()

    elif audio_type == 'str':

        if isinstance(final_audio, (torch.Tensor, np.ndarray)):
            if isinstance(final_audio, np.ndarray):
                final_audio = torch.from_numpy(final_audio)
            if final_audio.ndim < 2:
                final_audio = final_audio[None]
            torchaudio.save(temp_file, final_audio.cpu(), final_audio_sr)
            final_audio = temp_audio_file = temp_file

        elif isinstance(final_audio, bytes):
            with open(temp_file, 'wb') as f:
                f.write(final_audio)
            final_audio = temp_audio_file = temp_file

    elif audio_type == 'byte':

        if isinstance(final_audio, (torch.Tensor, np.ndarray)):
            if isinstance(final_audio, np.ndarray):
                final_audio = torch.from_numpy(final_audio)
            if final_audio.ndim < 2:
                final_audio = final_audio[None]
            with io.BytesIO() as f:
                torchaudio.save(f, final_audio.cpu(), final_audio_sr, format="wav")
                f.seek(0)
                final_audio = f.read()

        elif isinstance(final_audio, str):
            with open(final_audio, 'rb') as f:
                final_audio = f.read()

    inference_kwargs['audio'] = final_audio

    result = None
    try:
        result = inference_func(**inference_kwargs)
        if not isinstance(result, WhisperResult):
            result = WhisperResult(result, force_order=force_order, check_sorted=check_sorted)
        if suppress_silence:
            result.adjust_by_silence(
                audio, vad,
                vad_onnx=vad_onnx, vad_threshold=vad_threshold,
                q_levels=q_levels, k_size=k_size,
                sample_rate=curr_audio_sr(True), min_word_dur=min_word_dur,
                word_level=suppress_word_ts, verbose=True,
                nonspeech_error=nonspeech_error,
                use_word_position=use_word_position
            )

        if result.has_words and regroup:
            result.regroup(regroup)

    finally:
        if temp_audio_file is not None:
            try:
                os.unlink(temp_audio_file)
            except Exception as e:
                warnings.warn(f'Failed to remove temporary audio file {temp_audio_file}. {e}')

    return result



import subprocess
import time
from typing import List, Union, Tuple, Optional
from dataclasses import dataclass
from types import MethodType
import re

import numpy as np
import torch
from torch.nn.functional import pad

import whisper
from whisper.audio import (
    SAMPLE_RATE, N_FRAMES, HOP_LENGTH, N_SAMPLES, N_SAMPLES_PER_TOKEN, TOKENS_PER_SECOND, FRAMES_PER_SECOND, N_FFT,
    pad_or_trim, log_mel_spectrogram
)
from whisper.decoding import DecodingOptions

from .audio import AudioLoader
from .utils import isolate_useful_options
from .whisper_compatibility import get_tokenizer
from .decode import decode_stable
from .timing import _split_tokens, add_word_timestamps_stable
from .result import WhisperResult
from .default import get_append_punctuations, get_prepend_punctuations


@dataclass
class TempWord:
    word_dict: dict
    offset: Optional[float] = None

    def __post_init__(self):
        self.match = 0
        self._remove_appends = get_append_punctuations().replace('.', r'\.')
        self._remove_prepends = get_prepend_punctuations().replace('.', r'\.')

    @property
    def start(self) -> float:
        return self.word_dict['start']

    @property
    def end(self) -> float:
        return self.word_dict['end']

    @property
    def word(self) -> str:
        return self.word_dict['word']

    @property
    def probability(self) -> float:
        return self.word_dict['probability']

    @property
    def tokens(self) -> List[int]:
        return self.word_dict['tokens']

    @start.setter
    def start(self, new_start: float):
        self.word_dict['start'] = new_start

    @end.setter
    def end(self, new_end: float):
        self.word_dict['end'] = new_end

    def min_clamp(self, min_offset: float) -> bool:
        if self.start >= min_offset:
            return False
        self.start = min_offset
        if self.end >= min_offset:
            return False
        self.end = min_offset
        return True

    def update_word(self, new_word: "TempWord", update_ts: bool = False):
        if update_ts:
            # self.word_dict = new_word.word_dict
            # return
            if new_word.start < self.end:
                self.start = max(self.start, new_word.start)
            # self.end = min(self.end, new_word.end)
        self.word_dict['word'] = new_word.word
        self.word_dict['probability'] = new_word.probability
        self.word_dict['tokens'] = new_word.tokens

    def is_text_match(self, other_word: Union["TempWord", str]) -> bool:
        text = self.normalize_word()
        other_text = self.normalize_text(other_word) if isinstance(other_word, str) else other_word.normalize_word()
        return text == other_text[:len(text)] or text[-len(other_text):] == other_text

    def normalize_text(self, text: str) -> str:
        text = text.strip()
        text = re.sub(f'^[{self._remove_prepends}]', '', text)
        text = re.sub(f'[{self._remove_appends}]$', '', text)
        return text

    def normalize_word(self) -> str:
        return self.normalize_text(self.word)

    def apply_offset(self, offset: float):
        self.start += offset
        self.end += offset
        self.offset = offset
        return self


def fw_get_prompt(
        self,
        tokenizer,
        previous_tokens: List[int],
        without_timestamps: bool = False,
        prefix: Optional[Union[str, List[int]]] = None,
) -> List[int]:

    prompt = []

    if previous_tokens:
        prompt.append(tokenizer.sot_prev)
        prompt.extend(previous_tokens[-(self.max_length // 2 - 1) :])

    prompt.extend(tokenizer.sot_sequence)

    if without_timestamps:
        prompt.append(tokenizer.no_timestamps)

    if prefix:
        prefix_tokens = tokenizer.encode(" " + prefix.strip()) if isinstance(prefix, str) else prefix
        if len(prefix_tokens) >= self.max_length // 2:
            prefix_tokens = prefix_tokens[: self.max_length // 2 - 1]
        if not without_timestamps:
            prompt.append(tokenizer.timestamp_begin)
        prompt.extend(prefix_tokens)

    return prompt


class LiveWhisper:

    def __init__(self, model: whisper.Whisper, language: Optional[str] = None):
        if not isinstance(model, whisper.Whisper):
            raise NotImplementedError(f'Live transcription has not been implemented for {type(model)}.')
            # model.get_prompt = MethodType(fw_get_prompt, model)
        self._model = model
        self.dtype = torch.float16
        self._task = 'transcribe'
        self._language = language
        self._tokenizer = None
        self._prob_threshold = self._audio_feature = self._audio_chunk = self._chunk_mel = None
        self.unfinalized_words: List[TempWord] = []
        self.finalized_words: List[TempWord] = []
        self._min_prefix_index = 0
        self._seek = 0
        self._offset = 0.0

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, new_model):
        if new_model is not self._model:
            self._model = new_model

    @property
    def language(self):
        return self._language

    @language.setter
    def language(self, language: str):
        if self._language != language:
            self._language = language
            self._set_tokenizer()

    @property
    def prob_threshold(self):
        return self._prob_threshold

    @prob_threshold.setter
    def prob_threshold(self, threshold: float):
        self._prob_threshold = threshold

    @property
    def tokenizer(self):
        return self._tokenizer

    def _set_tokenizer(self):
        if self._tokenizer is None or self._tokenizer.language != self._language:
            self._tokenizer = get_tokenizer(self.model, task=self._task, language=self._language)

    def reset_states(self):
        self._prob_threshold = self._audio_feature = self._audio_chunk = self._chunk_mel = None
        self.unfinalized_words: List[TempWord] = []
        self.finalized_words: List[TempWord] = []
        self._min_prefix_index = 0
        self._seek = 0
        self._offset = 0.0

    def transcribe(
            self,
            audio,
            chunk_length: Union[int, str] = '0.25s',
            language: str = 'en',
            no_speech_threshold: float = 0.6,
            k=2,
            inference_chunk_size=5,
            prob_threshold: Optional[float] = 0.1,
            playback: bool = False,
            **kwargs
    ):
        if isinstance(chunk_length, str):
            assert chunk_length.endswith('s')
            chunk_length = round(float(chunk_length[:-1]) * SAMPLE_RATE)
        if not isinstance(audio, AudioLoader):
            audio = AudioLoader(audio, chunk_length, sr=SAMPLE_RATE, playback=playback)
        assert not isinstance(kwargs.get('temperature'), (list, tuple)), 'temperature can only be a single value'
        self.reset_states()
        self.language = language
        self.prob_threshold = prob_threshold

        next_seek = 0

        max_inference_samples = round(SAMPLE_RATE * inference_chunk_size)
        max_prefix_samples = max_inference_samples - chunk_length

        while True:

            prefix_tokens, prefix_words_count = self._get_prefix(next_seek, max_prefix_samples=max_prefix_samples)

            target_chunk_length = next_seek - self._seek + chunk_length
            next_seek = self._seek + target_chunk_length
            self._audio_chunk = audio.next_chunk(self._seek, target_chunk_length)
            if self._audio_chunk is None:
                break

            temp_words = self._transcribe_chunk(
                no_speech_threshold,
                prefix_tokens=prefix_tokens,
                prefix_words_count=prefix_words_count
            )
            if not temp_words:
                if prefix_tokens:
                    self._update_seek(next_seek)
                self._mark_nonspeech()
                continue

            last_i = self._update_unfinalized_words(temp_words, k)
            self.remove_unconfident_words(self.unfinalized_words)

            if target_chunk_length != self._audio_chunk.size(-1):
                last_i = len(self.unfinalized_words)

            self._update_finalized_words(last_i)

        audio.terminate()
        result = WhisperResult([[w.word_dict for w in self.finalized_words]])
        return result

    def _transcribe_chunk(
            self,
            no_speech_threshold: float,
            prefix_tokens: Optional[List[int]] = None,
            prefix_words_count: Optional[int] = None,
            **kwargs
    ) -> List[TempWord]:
        result = self._decode(prefix=prefix_tokens, **kwargs)
        if (not prefix_tokens and result.no_speech_prob > no_speech_threshold) or not result.tokens:
            return []

        return self._add_timestamp_words(
            tokens=result.tokens,
            prefix_tokens=prefix_tokens,
            prefix_words_count=prefix_words_count
        )

    def _update_unfinalized_words(self, temp_words: List[TempWord], k) -> int:
        last_i = None
        self.unfinalized_words = self.unfinalized_words[:len(temp_words)]
        min_offset = None
        for i, temp_word in enumerate(temp_words):
            if i < len(self.unfinalized_words):
                unfinalize_word = self.unfinalized_words[i]
                is_match = unfinalize_word.is_text_match(temp_word)
                if is_match:
                    unfinalize_word.update_word(temp_word, True)
                    unfinalize_word.match += 1
                    if unfinalize_word.match > k:
                        last_i = i + 1
                    continue
                else:
                    self.unfinalized_words = self.unfinalized_words[:i]
                    if self.unfinalized_words:
                        min_offset = self.unfinalized_words[-1].end

            if min_offset is not None and not temp_word.min_clamp(min_offset):
                min_offset = None

            self.unfinalized_words.append(temp_word)

        return last_i

    def _get_prefix(
            self, next_seek: int, max_prefix_samples: int
    ):
        prefix_count = None
        prefix_words = self.finalized_words[self._min_prefix_index:]
        if prefix_words:
            seek_time = round(max(next_seek - max_prefix_samples, self._seek) / SAMPLE_RATE, 3)
            for i, word in enumerate(reversed(prefix_words), 1):
                if word.start < seek_time:
                    break
                prefix_count = i

        if prefix_count is None:
            prefix_tokens = prefix_words_count = None
            self._update_seek(next_seek - max_prefix_samples)

            if self.unfinalized_words:
                for i, word in enumerate(reversed(self.unfinalized_words), 1):
                    if word.start < self._offset:
                        break
                    prefix_count = i
                finalize_count = (len(self.unfinalized_words) - prefix_count) if prefix_count else None
                self._update_finalized_words(finalize_count, update_seek=True)

        else:
            prefix = prefix_words[-prefix_count:]
            prefix_words_count = len(prefix)
            self._update_seek(round(prefix[0].start * SAMPLE_RATE), offset=prefix[0].start)
            prefix_tokens = [t for w in prefix for t in w.tokens]

        return prefix_tokens, prefix_words_count

    def _update_seek(self, seek: int, offset: Optional[float] = None, ignore_invalid: bool = True):
        if seek == self._seek or (seek < self._seek and ignore_invalid):
            return
        self._seek = seek
        if offset is None:
            offset = round(self._seek / SAMPLE_RATE, 3)
        self._offset = offset

    def on_words_callback(self, new_words: List[TempWord]):
        print(''.join(w.word for w in new_words))

    def _mark_nonspeech(self):
        self.unfinalized_words = []
        self._min_prefix_index = len(self.finalized_words)

    def _update_finalized_words(self, count: Optional[int], update_seek: bool = False):
        if not count or not self.unfinalized_words:
            return

        if self.finalized_words and self.unfinalized_words[0].is_text_match(self.finalized_words[-1]):
            self.finalized_words[-1].update_word(self.unfinalized_words[0], True)
            del self.unfinalized_words[0]
            if not self.unfinalized_words:
                return
            count -= 1

        new_words, self.unfinalized_words = self.unfinalized_words[:count], self.unfinalized_words[count:]
        if not new_words:
            return
        self.on_words_callback(new_words)
        self.finalized_words.extend(new_words)
        if update_seek:
            new_offset = self.finalized_words[-1].end
            self._update_seek(round(new_offset * SAMPLE_RATE), new_offset)

    def _decode(self, **kwargs):
        left_pad = N_SAMPLES - self._audio_chunk.size(-1)
        assert left_pad >= 0
        mel_chunk = log_mel_spectrogram(
            pad(self._audio_chunk, (0, left_pad)),
            self.model.dims.n_mels
        ).to(device=self.model.device, dtype=self.dtype)
        decoding_options = DecodingOptions(**kwargs, without_timestamps=True, language=self._language)
        result, self._audio_feature = decode_stable(self.model, mel_chunk, decoding_options, max_new_tokens=5)
        return result

    def _add_timestamp_words(
            self,
            tokens: List[int],
            prefix_tokens: Optional[List[int]],
            prefix_words_count: Optional[int],
    ) -> List[TempWord]:
        if prefix_tokens:
            tokens = prefix_tokens + tokens
        words, tokens = _split_tokens(tokens, self._tokenizer)

        segment = dict(
            seek=self._offset,
            tokens=(words, tokens)
        )

        add_word_timestamps_stable(
            segments=[segment],
            model=self._model,
            tokenizer=self._tokenizer,
            mel=self._chunk_mel,
            audio_features=self._audio_feature,
            num_samples=self._audio_chunk.size(-1),
            split_callback=(lambda x, _: x),
            gap_padding=None
        )

        words = segment['words'][prefix_words_count:] if prefix_words_count else segment['words']

        temp_words = [TempWord(word, offset=self._offset) for word in words]
        if prefix_tokens:
            self.min_clamp(temp_words)
        return temp_words

    def remove_unconfident_words(self, words: List[TempWord]):
        for i in reversed(range(len(words))):
            nonzero_duration = words[i].end - words[i].start != 0
            meet_prob_threshold = words[i].probability >= self._prob_threshold
            if nonzero_duration and meet_prob_threshold:
                break
            del words[i]

    def min_clamp(self, words: List[TempWord], min_offset: Optional[float] = None):
        if not words:
            return
        if min_offset is None and self.finalized_words:
            min_offset = self.finalized_words[-1].end
        if min_offset is None or words[0].start >= min_offset:
            return
        for w in words:
            if not w.min_clamp(min_offset):
                break
        self.remove_unconfident_words(words)



import warnings
from types import MethodType
from typing import TYPE_CHECKING, Union, Optional, Tuple, Callable

import torch
import numpy as np
from tqdm import tqdm

import whisper
from whisper import log_mel_spectrogram, pad_or_trim, DecodingResult, DecodingOptions
from whisper.audio import SAMPLE_RATE, N_FRAMES, HOP_LENGTH, N_SAMPLES_PER_TOKEN, N_SAMPLES
from whisper.tokenizer import LANGUAGES
from whisper.utils import exact_div

from ..result import WhisperResult, Segment
from ..audio import AudioLoader, audioloader_not_supported, convert_demucs_kwargs
from ..decode import decode_stable
from ..stabilization import NonSpeechPredictor
from ..timing import add_word_timestamps_stable
from ..utils import safe_print, isolate_useful_options, update_options
from ..whisper_compatibility import warn_compatibility_issues, get_tokenizer
from ..default import get_min_word_dur, get_prepend_punctuations, get_append_punctuations

if TYPE_CHECKING:
    from whisper.model import Whisper


def transcribe_stable(
        model: "Whisper",
        audio: Union[str, np.ndarray, torch.Tensor, bytes, AudioLoader],
        *,
        verbose: Optional[bool] = False,
        temperature: Union[float, Tuple[float, ...]] = (0.0, 0.2, 0.4, 0.6, 0.8, 1.0),
        compression_ratio_threshold: Optional[float] = 2.4,
        logprob_threshold: Optional[float] = -1.0,
        no_speech_threshold: Optional[float] = 0.6,
        condition_on_previous_text: bool = True,
        initial_prompt: Optional[str] = None,
        word_timestamps: bool = True,
        regroup: Union[bool, str] = True,
        ts_num: int = 0,
        ts_noise: float = 0.1,
        suppress_silence: bool = True,
        suppress_word_ts: bool = True,
        suppress_attention: bool = False,
        use_word_position: bool = True,
        q_levels: int = 20,
        k_size: int = 5,
        time_scale: float = None,
        denoiser: Optional[str] = None,
        denoiser_options: Optional[dict] = None,
        demucs: Union[bool, torch.nn.Module] = False,
        demucs_options: dict = None,
        vad: bool = False,
        vad_threshold: float = 0.35,
        vad_onnx: bool = False,
        min_word_dur: Optional[float] = None,
        nonspeech_error: float = 0.1,
        only_voice_freq: bool = False,
        prepend_punctuations: Optional[str] = None,
        append_punctuations: Optional[str] = None,
        stream: Optional[bool] = None,
        mel_first: Optional[bool] = None,
        split_callback: Callable = None,
        suppress_ts_tokens: bool = False,
        gap_padding: str = ' ...',
        only_ffmpeg: bool = False,
        max_instant_words: float = 0.5,
        avg_prob_threshold: Optional[float] = None,
        progress_callback: Callable = None,
        ignore_compatibility: bool = False,
        **decode_options) \
        -> WhisperResult:
    """
    Transcribe audio using Whisper.

    This is a modified version of :func:`whisper.transcribe.transcribe` with slightly different decoding logic while
    allowing additional preprocessing and postprocessing. The preprocessing performed on the audio includes:
    voice isolation / noise removal and low/high-pass filter. The postprocessing performed on the transcription
    result includes: adjusting timestamps with VAD and custom regrouping segments based punctuation and speech gaps.

    Parameters
    ----------
    model : whisper.model.Whisper
        An instance of Whisper ASR model.
    audio : str or numpy.ndarray or torch.Tensor or bytes or AudioLoader
        Path/URL to the audio file, the audio waveform, or bytes of audio file or
        instance of :class:`stable_whisper.audio.AudioLoader`.
        If audio is :class:`numpy.ndarray` or :class:`torch.Tensor`, the audio must be already at sampled to 16kHz.
    verbose : bool or None, default False
        Whether to display the text being decoded to the console.
        Displays all the details if ``True``. Displays progressbar if ``False``. Display nothing if ``None``.
    temperature : float or iterable of float, default (0.0, 0.2, 0.4, 0.6, 0.8, 1.0)
        Temperature for sampling. It can be a tuple of temperatures, which will be successfully used
        upon failures according to either ``compression_ratio_threshold`` or ``logprob_threshold``.
    compression_ratio_threshold : float, default 2.4
        If the gzip compression ratio is above this value, treat as failed.
    logprob_threshold : float, default -1
        If the average log probability over sampled tokens is below this value, treat as failed
    no_speech_threshold : float, default 0.6
        If the no_speech probability is higher than this value AND the average log probability
        over sampled tokens is below ``logprob_threshold``, consider the segment as silent
    condition_on_previous_text : bool, default True
        If ``True``, the previous output of the model is provided as a prompt for the next window;
        disabling may make the text inconsistent across windows, but the model becomes less prone to
        getting stuck in a failure loop, such as repetition looping or timestamps going out of sync.
    initial_prompt : str, optional
        Text to provide as a prompt for the first window. This can be used to provide, or
        "prompt-engineer" a context for transcription, e.g. custom vocabularies or proper nouns
        to make it more likely to predict those word correctly.
    word_timestamps : bool, default True
        Extract word-level timestamps using the cross-attention pattern and dynamic time warping,
        and include the timestamps for each word in each segment.
        Disabling this will prevent segments from splitting/merging properly.
    regroup : bool or str, default True, meaning the default regroup algorithm
        String for customizing the regrouping algorithm. False disables regrouping.
        Ignored if ``word_timestamps = False``.
    ts_num : int, default 0, meaning disable this option
        Number of extra timestamp inferences to perform then use average of these extra timestamps.
        An experimental option that might hurt performance.
    ts_noise : float, default 0.1
        Percentage of noise to add to audio_features to perform inferences for ``ts_num``.
    suppress_silence : bool, default True
        Whether to enable timestamps adjustments based on the detected silence.
    suppress_word_ts : bool, default True
        Whether to adjust word timestamps based on the detected silence. Only enabled if ``suppress_silence = True``.
    suppress_attention : bool, default False
        Whether to suppress cross-attention pattern with the predicted non-speech mask for computing word timestamps.
    use_word_position : bool, default True
        Whether to use position of the word in its segment to determine whether to keep end or start timestamps if
        adjustments are required. If it is the first word, keep end. Else if it is the last word, keep the start.
    q_levels : int, default 20
        Quantization levels for generating timestamp suppression mask; ignored if ``vad = true``.
        Acts as a threshold to marking sound as silent.
        Fewer levels will increase the threshold of volume at which to mark a sound as silent.
    k_size : int, default 5
        Kernel size for avg-pooling waveform to generate timestamp suppression mask; ignored if ``vad = true``.
        Recommend 5 or 3; higher sizes will reduce detection of silence.
    denoiser : str, optional
        String of the denoiser to use for preprocessing ``audio``.
        See ``stable_whisper.audio.SUPPORTED_DENOISERS`` for supported denoisers.
    denoiser_options : dict, optional
        Options to use for ``denoiser``.
    vad : bool, default False
        Whether to use Silero VAD to generate timestamp suppression mask.
        Silero VAD requires PyTorch 1.12.0+. Official repo, https://github.com/snakers4/silero-vad.
    vad_threshold : float, default 0.35
        Threshold for detecting speech with Silero VAD. Low threshold reduces false positives for silence detection.
    vad_onnx : bool, default False
        Whether to use ONNX for Silero VAD.
    min_word_dur : float or None, default None meaning use ``stable_whisper.default.DEFAULT_VALUES``
        Shortest duration each word is allowed to reach for silence suppression.
    nonspeech_error : float, default 0.1
        Relative error of non-speech sections that appear in between a word for silence suppression.
    only_voice_freq : bool, default False
        Whether to only use sound between 200 - 5000 Hz, where majority of human speech are.
    prepend_punctuations : str or None, default None meaning use ``stable_whisper.default.DEFAULT_VALUES``
        Punctuations to prepend to next word.
    append_punctuations : str or None, default None meaning use ``stable_whisper.default.DEFAULT_VALUES``
        Punctuations to append to previous word.
    stream : bool or None, default None
        Whether to loading ``audio`` in chunks of 30 seconds until the end of file/stream.
        If ``None`` and ``audio`` is a string then set to ``True`` else ``False``.
    mel_first : bool, optional
        Process entire audio track into log-Mel spectrogram first instead in chunks.
        Used if odd behavior seen in stable-ts but not in whisper, but use significantly more memory for long audio.
    split_callback : Callable, optional
        Custom callback for grouping tokens up with their corresponding words.
        The callback must take two arguments, list of tokens and tokenizer.
        The callback returns a tuple with a list of words and a corresponding nested list of tokens.
    suppress_ts_tokens : bool, default False
        Whether to suppress timestamp tokens during inference for timestamps are detected at silent.
        Reduces hallucinations in some cases, but also prone to ignore disfluencies and repetitions.
        This option is ignored if ``suppress_silence = False``.
    gap_padding : str, default ' ...'
        Padding prepend to each segments for word timing alignment.
        Used to reduce the probability of model predicting timestamps earlier than the first utterance.
    only_ffmpeg : bool, default False
        Whether to use only FFmpeg (instead of not yt-dlp) for URls
    max_instant_words : float, default 0.5
        If percentage of instantaneous words in a segment exceed this amount, the segment is removed.
    avg_prob_threshold: float or None, default None
        Transcribe the gap after the previous word and if the average word proababiliy of a segment falls below this
        value, discard the segment. If ``None``, skip transcribing the gap to reduce chance of timestamps starting
        before the next utterance.
    progress_callback : Callable, optional
        A function that will be called when transcription progress is updated.
        The callback need two parameters.
        The first parameter is a float for seconds of the audio that has been transcribed.
        The second parameter is a float for total duration of audio in seconds.
    ignore_compatibility : bool, default False
        Whether to ignore warnings for compatibility issues with the detected Whisper version.
    decode_options
        Keyword arguments to construct class:`whisper.decode.DecodingOptions` instances.

    Returns
    -------
    stable_whisper.result.WhisperResult
        All timestamps, words, probabilities, and other data from the transcription of ``audio``.

    See Also
    --------
    stable_whisper.non_whisper.transcribe_any : Return :class:`stable_whisper.result.WhisperResult` containing all the
        data from transcribing audio with unmodified :func:`whisper.transcribe.transcribe` with preprocessing and
        postprocessing.
    stable_whisper.whisper_word_level.faster_whisper.faster_transcribe : Return
        :class:`stable_whisper.result.WhisperResult` containing all the data from transcribing audio with
        :meth:`faster_whisper.WhisperModel.transcribe` with preprocessing and postprocessing.

    Examples
    --------
    >>> import stable_whisper
    >>> model = stable_whisper.load_model('base')
    >>> result = model.transcribe('audio.mp3', vad=True)
    >>> result.to_srt_vtt('audio.srt')
    Saved: audio.srt
    """
    if mel_first is not None:
        warnings.warn('``mel_first`` is deprecated and will be removed in future versions.'
                      'Use ``stream`` (e.g. replace ``mel_first=True`` with ``stream=False``).',
                      stacklevel=2)
        stream = not mel_first

    prepend_punctuations = get_prepend_punctuations(prepend_punctuations)
    append_punctuations = get_append_punctuations(append_punctuations)
    min_word_dur = get_min_word_dur(min_word_dur)

    warn_compatibility_issues(whisper, ignore_compatibility, 'Or use transcribe_minimal().')
    dtype = torch.float16 if decode_options.get("fp16", True) and not getattr(model, 'dq', False) else torch.float32
    if model.device == torch.device("cpu"):
        if torch.cuda.is_available():
            warnings.warn("Performing inference on CPU when CUDA is available")
        if dtype == torch.float16:
            warnings.warn("FP16 is not supported on CPU; using FP32 instead")
            dtype = torch.float32

    if dtype == torch.float32:
        decode_options["fp16"] = False

    if 'max_initial_timestamp' not in decode_options:
        decode_options['max_initial_timestamp'] = None

    device = model.device

    if time_scale:
        warnings.warn('``time_scale`` is deprecated and will be removed in future versions. '
                      'It currently does not affect results.',
                      stacklevel=2)
    if decode_options.pop('input_sr', None):
        warnings.warn('``input_sr`` is deprecated and will be removed in future versions. '
                      '``audio`` of types numpy.ndarray and torch.Tensor inputs must be already at 16kHz. '
                      'To higher sample rates for ``audio`` use str or bytes.',
                      stacklevel=2)
    denoiser, denoiser_options = convert_demucs_kwargs(
        denoiser, denoiser_options, demucs=demucs, demucs_options=demucs_options
    )

    if isinstance(audio, AudioLoader):
        audio.validate_external_args(
            sr=SAMPLE_RATE,
            vad=vad,
            stream=stream,
            denoiser=denoiser,
            denoiser_options=denoiser_options,
            only_voice_freq=only_voice_freq
        )
    else:
        denoiser_options = update_options(denoiser_options, device=device)
        audio = AudioLoader(
            audio,
            stream=stream,
            denoiser=denoiser,
            denoiser_options=denoiser_options,
            only_voice_freq=only_voice_freq,
            only_ffmpeg=only_ffmpeg,
            verbose=verbose,
            new_chunk_divisor=512 if vad else None
        )
    tokenizer = None
    language = None
    initial_prompt_tokens = []
    task = decode_options.get("task", "transcribe")
    if word_timestamps and task == "translate":
        warnings.warn("Word-level timestamps on translations may not be reliable.")

    def detect_language():
        nonlocal tokenizer
        if tokenizer is None:
            if not decode_options.get("language"):
                if not model.is_multilingual:
                    decode_options["language"] = "en"
                else:
                    if verbose:
                        print("Detecting language using up to 30 seconds following first non-silent sample. "
                              "Use `--language` to specify the language")
                    _, probs = model.detect_language(mel_segment)
                    decode_options["language"] = max(probs, key=probs.get)
                    if verbose is not None:
                        detected_msg = f"Detected language: {LANGUAGES[decode_options['language']]}"
                        if tqdm_pbar.disable:
                            print(detected_msg)
                        else:
                            tqdm_pbar.write(detected_msg)

            nonlocal language
            language = decode_options["language"]
            tokenizer = get_tokenizer(model, language=language, task=task)

            if initial_prompt is not None:
                nonlocal initial_prompt_tokens
                initial_prompt_tokens = tokenizer.encode(" " + initial_prompt.strip())
                all_tokens.extend(initial_prompt_tokens)

    audio_features = None

    def decode_with_fallback(seg: torch.Tensor,
                             ts_token_mask: torch.Tensor = None) \
            -> DecodingResult:
        nonlocal audio_features
        temperatures = [temperature] if isinstance(temperature, (int, float)) else temperature
        decode_result = None

        for t in temperatures:
            kwargs = {**decode_options}
            if t > 0:
                # disable beam_size and patience when t > 0
                kwargs.pop("beam_size", None)
                kwargs.pop("patience", None)
            else:
                # disable best_of when t == 0
                kwargs.pop("best_of", None)

            options = DecodingOptions(**kwargs, temperature=t)
            decode_result, audio_features = decode_stable(model,
                                                          seg,
                                                          options,
                                                          ts_token_mask=ts_token_mask if suppress_ts_tokens else None,
                                                          audio_features=audio_features)

            needs_fallback = False
            if (
                    compression_ratio_threshold is not None
                    and decode_result.compression_ratio > compression_ratio_threshold
            ):
                needs_fallback = True  # too repetitive
            if (
                    logprob_threshold is not None
                    and decode_result.avg_logprob < logprob_threshold
            ):
                needs_fallback = True  # average log probability is too low
            if (
                no_speech_threshold is not None
                and decode_result.no_speech_prob > no_speech_threshold
            ):
                needs_fallback = False  # silence

            if not needs_fallback:
                break

        return decode_result

    seek_sample = 0  # samples
    input_stride = exact_div(
        N_FRAMES, model.dims.n_audio_ctx
    )  # mel frames per output token: 2
    time_precision = (
            input_stride * HOP_LENGTH / SAMPLE_RATE
    )  # time per output token: 0.02 (seconds)
    all_tokens = []
    all_segments = []
    prompt_reset_since = 0

    def new_segment(
            *, start: float, end: float, tokens: torch.Tensor, result: DecodingResult
    ):
        tokens = tokens.tolist()
        text_tokens = [token for token in tokens if token < tokenizer.eot]
        return {
            "seek": round(seek_sample / SAMPLE_RATE, 3),  # units in seconds
            "start": start,
            "end": end,
            "text": tokenizer.decode(text_tokens),
            "tokens": tokens,
            "temperature": result.temperature,
            "avg_logprob": result.avg_logprob,
            "compression_ratio": result.compression_ratio,
            "no_speech_prob": result.no_speech_prob,
        }

    punctuations = prepend_punctuations + append_punctuations

    initial_duration = audio.get_duration(2)

    nonspeech_predictor = NonSpeechPredictor(
        vad=vad if (suppress_silence or suppress_attention) else None,
        mask_pad_func=pad_or_trim,
        get_mask=suppress_ts_tokens or suppress_attention,
        min_word_dur=min_word_dur,
        q_levels=q_levels,
        k_size=k_size,
        vad_threshold=vad_threshold,
        vad_onnx=vad_onnx,
        vad_window=512,
        sampling_rate=SAMPLE_RATE,
        verbose=None if audio.stream else verbose,
        store_timings=True
    )
    audio.update_post_prep_callback(nonspeech_predictor.get_on_prep_callback(audio.stream))

    with tqdm(total=initial_duration, unit='sec', disable=verbose is not False, desc=task.title()) as tqdm_pbar:

        def update_pbar():
            nonlocal audio_features
            audio_features = None
            curr_total_duration = audio.get_duration(2)
            if curr_total_duration != tqdm_pbar.total:
                tqdm_pbar.total = curr_total_duration
                tqdm_pbar.refresh()
            seek_duration = min(curr_total_duration, round(seek_sample / SAMPLE_RATE, 2))
            if not tqdm_pbar.disable:
                tqdm_pbar.update(seek_duration - tqdm_pbar.n)
            if progress_callback is not None:
                progress_callback(seek=seek_duration, total=curr_total_duration)

        def update_seek():
            nonlocal seek_sample
            seek_sample += segment_samples

        def fast_forward():
            # fast-forward to the next segment boundary
            update_seek()
            update_pbar()

        while True:
            audio_segment = audio.next_chunk(seek_sample, N_SAMPLES)
            if audio_segment is None:
                break
            time_offset = seek_sample / SAMPLE_RATE
            segment_samples = audio_segment.shape[-1]
            segment_duration = segment_samples / SAMPLE_RATE

            silence_preds = nonspeech_predictor.predict(audio_segment, offset=time_offset)
            segment_silence_timing = silence_preds['timings'] if suppress_silence else None
            ts_token_mask = silence_preds['mask'] if suppress_ts_tokens else None
            is_silent_segment = silence_preds['is_silent']

            if is_silent_segment:
                fast_forward()
                continue

            sample_padding = max(N_SAMPLES - segment_samples, 0)
            mel_segment = log_mel_spectrogram(audio_segment, model.dims.n_mels, padding=sample_padding)
            mel_segment = pad_or_trim(mel_segment, N_FRAMES).to(device=model.device, dtype=dtype)

            detect_language()
            decode_options["prompt"] = all_tokens[prompt_reset_since:]
            result: DecodingResult = decode_with_fallback(mel_segment, ts_token_mask=ts_token_mask)
            tokens = torch.tensor(result.tokens)

            if no_speech_threshold is not None:
                # no voice activity check
                should_skip = result.no_speech_prob > no_speech_threshold
                if logprob_threshold is not None and result.avg_logprob > logprob_threshold:
                    # don't skip if the logprob is high enough, despite the no_speech_prob
                    should_skip = False

                if should_skip:
                    fast_forward()
                    continue

            current_segments = []

            timestamp_tokens: torch.Tensor = tokens.ge(tokenizer.timestamp_begin)
            single_timestamp_ending = timestamp_tokens[-2:].tolist() == [False, True]

            consecutive = torch.where(timestamp_tokens[:-1] & timestamp_tokens[1:])[0]
            consecutive.add_(1)
            if len(consecutive) > 0:
                # if the output contains two consecutive timestamp tokens
                slices = consecutive.tolist()
                if single_timestamp_ending:
                    slices.append(len(tokens))

                last_slice = 0
                for current_slice in slices:
                    sliced_tokens = tokens[last_slice:current_slice]
                    start_timestamp_pos = (
                            sliced_tokens[0].item() - tokenizer.timestamp_begin
                    )
                    end_timestamp_pos = (
                            sliced_tokens[-1].item() - tokenizer.timestamp_begin
                    )
                    current_segments.append(
                        new_segment(
                            start=round(time_offset + start_timestamp_pos * time_precision, 3),
                            end=round(time_offset + min(end_timestamp_pos * time_precision, segment_duration), 3),
                            tokens=sliced_tokens,
                            result=result,
                        )
                    )
                    last_slice = current_slice

            else:
                duration = segment_duration
                timestamps = tokens[timestamp_tokens.nonzero().flatten()]
                if (
                        len(timestamps) > 0
                        and timestamps[-1].item() != tokenizer.timestamp_begin
                ):
                    # no consecutive timestamps but it has a timestamp; use the last one.
                    end_timestamp_pos = (
                            timestamps[-1].item() - tokenizer.timestamp_begin
                    )
                    duration = min(end_timestamp_pos * time_precision, segment_duration)
                else:
                    end_timestamp_pos = 0

                current_segments.append(
                    new_segment(
                        start=round(time_offset, 3),
                        end=round(time_offset + duration, 3),
                        tokens=tokens,
                        result=result,
                    )
                )

            for i in reversed(range(len(current_segments))):
                seg = current_segments[i]
                if seg["text"].strip() in punctuations:
                    del current_segments[i]
                else:
                    if word_timestamps:
                        if seg["start"] == seg["end"]:
                            del current_segments[i]
                    else:
                        prev_i = i+1
                        if prev_i >= len(current_segments):
                            max_end = seg['end']
                        else:
                            max_end = current_segments[prev_i]['start']
                        if seg["start"] > seg["end"]:
                            if (
                                    i != 0 and
                                    current_segments[i - 1]['end'] != current_segments[i - 1]['start'] and
                                    current_segments[i - 1]['end'] < max_end
                            ):
                                new_start = current_segments[i-1]['end']
                            else:
                                new_start = max_end
                            seg['start'] = new_start

            num_samples = (
                min(round(end_timestamp_pos * N_SAMPLES_PER_TOKEN), segment_samples)
                if end_timestamp_pos > 0 else
                segment_samples
            )

            if word_timestamps:
                add_word_timestamps_stable(
                    segments=current_segments,
                    model=model,
                    tokenizer=tokenizer,
                    mel=mel_segment,
                    num_samples=num_samples,
                    prepend_punctuations=prepend_punctuations,
                    append_punctuations=append_punctuations,
                    audio_features=audio_features,
                    ts_num=ts_num,
                    ts_noise=ts_noise,
                    split_callback=split_callback,
                    gap_padding=gap_padding,
                    ts_token_mask=(silence_preds['mask'] if suppress_attention else None)
                )

                for i in reversed(range(len(current_segments))):
                    zero_duration_percent = (
                        np.array(
                            [w['start'] == w['end'] for w in current_segments[i]['words']]
                        )
                        .astype(np.float16)
                        .mean()
                    )
                    if zero_duration_percent > max_instant_words:
                        del current_segments[i]

                if avg_prob_threshold and current_segments:
                    if (
                            single_timestamp_ending and
                            (np.mean([w['probability'] for s in current_segments for w in s['words']]) <
                             avg_prob_threshold)
                    ):
                        num_samples = segment_samples
                        current_segments = []
                    else:
                        num_samples = round((current_segments[-1]['words'][-1]['end']-time_offset) * SAMPLE_RATE)

            if len(current_segments) == 0:
                fast_forward()
                continue

            if segment_silence_timing is not None:
                for seg_i, segment in enumerate(current_segments):
                    segment = Segment(**segment, ignore_unused_args=True).suppress_silence(
                            *segment_silence_timing,
                            min_word_dur=min_word_dur,
                            word_level=suppress_word_ts,
                            nonspeech_error=nonspeech_error,
                            use_word_position=use_word_position,
                        )
                    if verbose:
                        safe_print(segment.to_display_str())
                    current_segments[seg_i] = segment.to_dict()

            all_segments.extend(
                [
                    {"id": i, **segment}
                    for i, segment in enumerate(current_segments, start=len(all_segments))
                ]
            )
            all_tokens.extend(
                [token for segment in current_segments for token in segment["tokens"]]
            )
            if not single_timestamp_ending or avg_prob_threshold:
                segment_samples = num_samples

            if not condition_on_previous_text or result.temperature > 0.5:
                # do not feed the prompt tokens if a high temperature was used
                prompt_reset_since = len(all_tokens)

            fast_forward()

        # final update
        update_pbar()

    if model.device != torch.device('cpu'):
        torch.cuda.empty_cache()

    audio.terminate()
    nonspeech_predictor.finalize_timings()

    text = '' if tokenizer is None else tokenizer.decode(all_tokens[len(initial_prompt_tokens):])
    final_result = WhisperResult(
        dict(
            text=text,
            segments=all_segments,
            language=language,
            time_scale=time_scale
        ),
        force_order=not word_timestamps
    )
    if word_timestamps and regroup:
        final_result.regroup(regroup)

    if time_scale is not None:
        final_result.rescale_time(1 / time_scale)

    if len(final_result.text) == 0:
        warnings.warn(f'Failed to {task} audio. Result contains no text. ')

    if suppress_silence and (final_nonspeech_timings := nonspeech_predictor.nonspeech_timings):
        final_result.update_nonspeech_sections(*final_nonspeech_timings)

    return final_result


def transcribe_minimal(
        model: "Whisper",
        audio: Union[str, np.ndarray, torch.Tensor, bytes],
        *,
        verbose: Optional[bool] = False,
        word_timestamps: bool = True,
        regroup: Union[bool, str] = True,
        suppress_silence: bool = True,
        suppress_word_ts: bool = True,
        use_word_position: bool = True,
        q_levels: int = 20,
        k_size: int = 5,
        denoiser: Optional[str] = None,
        denoiser_options: Optional[dict] = None,
        demucs: bool = False,
        demucs_options: dict = None,
        vad: bool = False,
        vad_threshold: float = 0.35,
        vad_onnx: bool = False,
        min_word_dur: float = 0.1,
        nonspeech_error: float = 0.1,
        only_voice_freq: bool = False,
        only_ffmpeg: bool = False,
        **options) \
        -> WhisperResult:
    """
    Transcribe audio using Whisper.

    This is uses the original whisper transcribe function, :func:`whisper.transcribe.transcribe`, while still allowing
    additional preprocessing and postprocessing. The preprocessing performed on the audio includes: voice isolation /
    noise removal and low/high-pass filter. The postprocessing performed on the transcription result includes:
    adjusting timestamps with VAD and custom regrouping segments based punctuation and speech gaps.

    Parameters
    ----------
    model : whisper.model.Whisper
        An instance of Whisper ASR model.
    audio : str or numpy.ndarray or torch.Tensor or bytes
        Path/URL to the audio file, the audio waveform, or bytes of audio file.
        If audio is ``numpy.ndarray`` or ``torch.Tensor``, the audio must be already at sampled to 16kHz.
    verbose : bool or None, default False
        Whether to display the text being decoded to the console.
        Displays all the details if ``True``. Displays progressbar if ``False``. Display nothing if ``None``.
    word_timestamps : bool, default True
        Extract word-level timestamps using the cross-attention pattern and dynamic time warping,
        and include the timestamps for each word in each segment.
        Disabling this will prevent segments from splitting/merging properly.
    regroup : bool or str, default True, meaning the default regroup algorithm
        String for customizing the regrouping algorithm. False disables regrouping.
        Ignored if ``word_timestamps = False``.
    suppress_silence : bool, default True
        Whether to enable timestamps adjustments based on the detected silence.
    suppress_word_ts : bool, default True
        Whether to adjust word timestamps based on the detected silence. Only enabled if ``suppress_silence = True``.
    use_word_position : bool, default True
        Whether to use position of the word in its segment to determine whether to keep end or start timestamps if
        adjustments are required. If it is the first word, keep end. Else if it is the last word, keep the start.
    q_levels : int, default 20
        Quantization levels for generating timestamp suppression mask; ignored if ``vad = true``.
        Acts as a threshold to marking sound as silent.
        Fewer levels will increase the threshold of volume at which to mark a sound as silent.
    k_size : int, default 5
        Kernel size for avg-pooling waveform to generate timestamp suppression mask; ignored if ``vad = true``.
        Recommend 5 or 3; higher sizes will reduce detection of silence.
    denoiser : str, optional
        String of the denoiser to use for preprocessing ``audio``.
        See ``stable_whisper.audio.SUPPORTED_DENOISERS`` for supported denoisers.
    denoiser_options : dict, optional
        Options to use for ``denoiser``.
    vad : bool, default False
        Whether to use Silero VAD to generate timestamp suppression mask.
        Silero VAD requires PyTorch 1.12.0+. Official repo, https://github.com/snakers4/silero-vad.
    vad_threshold : float, default 0.35
        Threshold for detecting speech with Silero VAD. Low threshold reduces false positives for silence detection.
    vad_onnx : bool, default False
        Whether to use ONNX for Silero VAD.
    min_word_dur : float, default 0.1
        Shortest duration each word is allowed to reach for silence suppression.
    nonspeech_error : float, default 0.1
        Relative error of non-speech sections that appear in between a word for silence suppression.
    only_voice_freq : bool, default False
        Whether to only use sound between 200 - 5000 Hz, where majority of human speech are.
    only_ffmpeg : bool, default False
        Whether to use only FFmpeg (instead of not yt-dlp) for URls
    options
        Additional options used for :func:`whisper.transcribe.transcribe` and
        :func:`stable_whisper.non_whisper.transcribe_any`.
    Returns
    -------
    stable_whisper.result.WhisperResult
        All timestamps, words, probabilities, and other data from the transcription of ``audio``.

    Examples
    --------
    >>> import stable_whisper
    >>> model = stable_whisper.load_model('base')
    >>> result = model.transcribe_minimal('audio.mp3', vad=True)
    >>> result.to_srt_vtt('audio.srt')
    Saved: audio.srt
    """
    audioloader_not_supported(audio)
    from ..non_whisper import transcribe_any
    inference_kwargs = dict(
        model=model,
        audio=audio,
        word_timestamps=word_timestamps,
        verbose=verbose
    )
    extra_options = isolate_useful_options(options, transcribe_any, True)
    denoiser, denoiser_options = convert_demucs_kwargs(
        denoiser, denoiser_options, demucs=demucs, demucs_options=demucs_options
    )
    if not isinstance(audio, (str, bytes)):
        if 'input_sr' not in extra_options:
            extra_options['input_sr'] = SAMPLE_RATE

    if denoiser or only_voice_freq:
        if 'audio_type' not in extra_options:
            extra_options['audio_type'] = 'torch'
        if 'model_sr' not in extra_options:
            extra_options['model_sr'] = SAMPLE_RATE
    inference_kwargs.update(options)
    return transcribe_any(
        inference_func=whisper.transcribe,
        audio=audio,
        inference_kwargs=inference_kwargs,
        verbose=verbose,
        regroup=regroup,
        suppress_silence=suppress_silence,
        suppress_word_ts=suppress_word_ts,
        q_levels=q_levels,
        k_size=k_size,
        denoiser=denoiser,
        denoiser_options=denoiser_options,
        vad=vad,
        vad_threshold=vad_threshold,
        vad_onnx=vad_onnx,
        min_word_dur=min_word_dur,
        nonspeech_error=nonspeech_error,
        use_word_position=use_word_position,
        only_voice_freq=only_voice_freq,
        only_ffmpeg=only_ffmpeg,
        force_order=True,
        **extra_options
    )


def modify_model(model: "Whisper"):
    """
    Modify an instance if :class:`whisper.model.Whisper`.

    The following are performed:
    -replace :meth:`whisper.model.Whisper.transcribe` with :func:`stable_whisper.whisper_word_level.transcribe_stable`
    -assign :meth:`whisper.model.transcribe_minimal` to :func:`stable_whisper.whisper_word_level.transcribe_minimal`
    -assign :meth:`whisper.model.Whisper.transcribe_original` to :meth:`whisper.model.Whisper.transcribe`
    -assign :meth:`whisper.model.Whisper.align` to :func:`stable_whisper.alignment.align`
    -assign :meth:`whisper.model.Whisper.locate` to :func:`stable_whisper.alignment.locate`
    """
    model.transcribe = MethodType(transcribe_stable, model)
    model.transcribe_minimal = MethodType(transcribe_minimal, model)
    model.transcribe_original = MethodType(whisper.transcribe, model)
    from ..alignment import align, refine, locate
    model.align = MethodType(align, model)
    model.refine = MethodType(refine, model)
    model.locate = MethodType(locate, model)


# modified version of whisper.load_model
def load_model(name: str, device: Optional[Union[str, torch.device]] = None,
               download_root: str = None, in_memory: bool = False,
               cpu_preload: bool = True, dq: bool = False) -> "Whisper":
    """
    Load an instance if :class:`whisper.model.Whisper`.

    Parameters
    ----------
    name : {'tiny', 'tiny.en', 'base', 'base.en', 'small', 'small.en', 'medium', 'medium.en', 'large-v1',
        'large-v2', 'large-v3', or 'large'}
        One of the official model names listed by :func:`whisper.available_models`, or
        path to a model checkpoint containing the model dimensions and the model state_dict.
    device : str or torch.device, optional
        PyTorch device to put the model into.
    download_root : str, optional
        Path to download the model files; by default, it uses "~/.cache/whisper".
    in_memory : bool, default False
        Whether to preload the model weights into host memory.
    cpu_preload : bool, default True
        Load model into CPU memory first then move model to specified device
        to reduce GPU memory usage when loading model
    dq : bool, default False
        Whether to apply Dynamic Quantization to model to reduced memory usage and increase inference speed
        but at the cost of a slight decrease in accuracy. Only for CPU.

    Returns
    -------
    model : "Whisper"
        The Whisper ASR model instance.

    Notes
    -----
    The overhead from ``dq = True`` might make inference slower for models smaller than 'large'.
    """
    if device is None or dq:
        device = "cuda" if torch.cuda.is_available() and not dq else "cpu"
    if cpu_preload:
        model = whisper.load_model(name, device='cpu', download_root=download_root, in_memory=in_memory)
        cuda_index = None
        if isinstance(device, str) and device.startswith('cuda'):
            try:
                cuda_index = [] if device == 'cuda' else [int(device.split(':')[-1])]
            except ValueError:
                pass
        model = model.to(device=device) if cuda_index is None else model.cuda(*cuda_index)
    else:
        model = whisper.load_model(name, device=device, download_root=download_root, in_memory=in_memory)
    modify_model(model)
    if dq:
        from ..quantization import ptdq_linear
        ptdq_linear(model)
    return model



import os
import warnings
import argparse
from typing import Optional, List, Union, Dict
from os.path import splitext, split, join
import shlex

import numpy as np
import torch

from ..result import WhisperResult
from ..utils import isolate_useful_options, str_to_valid_type, get_func_parameters
from ..audio import SUPPORTED_DENOISERS
from ..default import *

from whisper import DecodingOptions, available_models
from whisper.tokenizer import LANGUAGES, TO_LANGUAGE_CODE
from whisper.utils import optional_int, optional_float


def _split_input_args(args: str) -> List[str]:
    args = shlex.split(args)
    if args and args[0] == 'stable-ts':
        del args[0]
    return args


# modified version of whisper.transcribe.cli
def _cli(cmd: str = None, _cache: Dict[str, Union[bool, dict]] = None):

    supported_denoisers = tuple(SUPPORTED_DENOISERS.keys())

    str2val = {"true": True, "false": False, "1": True, "0": False}

    def str2bool(string: str) -> bool:
        string = string.lower()
        if string in str2val:
            return str2val[string]
        raise ValueError(f"Expected one of {set(str2val.keys())}, got {string}")

    def valid_model_name(name):
        if name in available_models() or os.path.exists(name):
            return name
        raise ValueError(
            f"model should be one of {available_models()} or path to a model checkpoint"
        )

    def use_deprecated_args(
            key: str, old_key: str, pop: bool = False, expected_default=None, new_default=None, eg: str = None,
    ):
        new_val = args.pop(key) if pop else args[key]
        old_val = args.pop(old_key) if pop else args[old_key]

        if old_val != expected_default:
            eg_str = eg if eg is None else f' (e.g. {eg})'
            warnings.warn(f'{old_key} is deprecated and will be removed in future versions. '
                          f'Use {key}{eg_str}.', stacklevel=2)
            if new_val == expected_default:
                new_val = old_val
        elif new_val == expected_default:
            new_val = new_default
        return new_val

    def update_options_with_args(arg_key: Union[str, list], options: Optional[dict] = None, pop: bool = False):
        extra_options = arg_key if isinstance(arg_key, list) else (args.pop(arg_key) if pop else args.get(arg_key))
        if not extra_options:
            return
        extra_options = [kv.split('=', maxsplit=1) for kv in extra_options]
        missing_val = [kv[0] for kv in extra_options if len(kv) == 1]
        if missing_val:
            raise ValueError(f'Following expected values for the following custom options: {missing_val}')
        extra_options = dict((k, str_to_valid_type(v)) for k, v in extra_options)
        if options is None:
            return extra_options
        options.update(extra_options)

    def url_to_path(url: str):
        if '://' in url:
            from urllib.parse import urlparse
            return urlparse(url).path.strip('/')
        return url

    OUTPUT_FORMATS_METHODS = {
        "srt": "to_srt_vtt",
        "ass": "to_ass",
        "json": "save_as_json",
        "vtt": "to_srt_vtt",
        "tsv": "to_tsv",
        "txt": "to_txt",
    }

    OUTPUT_FORMATS = set(OUTPUT_FORMATS_METHODS.keys())

    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("inputs", nargs="+", type=str,
                        help="audio/video filepath/URL(s) to transcribe "
                             "or json file(s) to process into [output_format]")
    parser.add_argument("--output", "-o", action="extend", nargs="+", type=str,
                        help="output filepaths(s);"
                             "if not specified, auto-named output file(s) will be saved to "
                             "[output_dir] or current dir if not specified.")
    parser.add_argument("--model", '-m', default="base", type=valid_model_name,
                        help="name of the Whisper model to use")
    parser.add_argument("--model_dir", type=str, default=None,
                        help="the path to save model files; uses ~/.cache/whisper by default")
    parser.add_argument("--device", default="cuda" if torch.cuda.is_available() else "cpu",
                        help="device to use for PyTorch inference")
    parser.add_argument("--cpu_preload", type=str2bool, default=True,
                        help="load model into CPU memory first then move model to specified device; "
                             "this reduces GPU memory usage when loading model.")
    parser.add_argument("--output_dir", "-d", type=str,
                        help="directory to save the outputs;"
                             "if a path in [output] does not have parent, that output will be save to this directory")
    parser.add_argument("--output_format", "-f", type=str,
                        help="format of the output file(s); "
                             f"Supported Formats: {OUTPUT_FORMATS}; "
                             "use ',' to separate multiple formats")
    parser.add_argument("--verbose", '-v', type=int, default=1, choices=(0, 1, 2),
                        help="whether to display the text being decoded to the console; "
                             "if 2, display all the details; "
                             "if 1, display progressbar; "
                             "if 0, display nothing")

    parser.add_argument("--dynamic_quantization", "-dq", action='store_true',
                        help="whether to apply Dynamic Quantization to model "
                             "to reduced memory usage (~half less) and increase inference speed "
                             "at cost of slight decrease in accuracy; Only for CPU; "
                             "NOTE: overhead might make inference slower for models smaller than 'large'")

    parser.add_argument("--task", type=str, default="transcribe",
                        choices=["transcribe", "translate"],
                        help="whether to perform X->X speech recognition ('transcribe') "
                             "or X->English translation ('translate')")
    parser.add_argument("--language", '-l', type=str, default=None,
                        choices=sorted(LANGUAGES.keys()) + sorted([k.title() for k in TO_LANGUAGE_CODE.keys()]),
                        help="language spoken in the audio, specify None to perform language detection")

    parser.add_argument("--prepend_punctuations", '-pp', type=str, default=get_prepend_punctuations(),
                        help="Punctuations to prepend to next word")
    parser.add_argument("--append_punctuations", '-ap', type=str, default=get_append_punctuations(),
                        help="Punctuations to append to previous word")

    parser.add_argument("--gap_padding", type=str, default=" ...",
                        help="padding prepend to each segments for word timing alignment;"
                             "used to reduce the probability of model predicting timestamps "
                             "earlier than the first utterance")

    parser.add_argument("--word_timestamps", type=str2bool, default=True,
                        help="extract word-level timestamps using the cross-attention pattern and dynamic time warping,"
                             "and include the timestamps for each word in each segment;"
                             "disabling this will prevent segments from splitting/merging properly.")

    parser.add_argument("--regroup", type=str, default="True",
                        help="whether to regroup all words into segments with more natural boundaries;"
                             "specify string for customizing the regrouping algorithm"
                             "ignored if [word_timestamps]=False.")

    parser.add_argument('--ts_num', type=int, default=0,
                        help="number of extra inferences to perform to find the mean timestamps")
    parser.add_argument('--ts_noise', type=float, default=0.1,
                        help="percentage of noise to add to audio_features to perform inferences for [ts_num]")

    parser.add_argument('--suppress_silence', type=str2bool, default=True,
                        help="whether to suppress timestamp where audio is silent at segment-level"
                             "and word-level if [suppress_word_ts]=True")
    parser.add_argument('--suppress_word_ts', type=str2bool, default=True,
                        help="whether to suppress timestamps where audio is silent at word-level; "
                             "ignored if [suppress_silence]=False")

    parser.add_argument('--suppress_ts_tokens', type=str2bool, default=False,
                        help="whether to use silence mask to suppress silent timestamp tokens during inference; "
                             "increases word accuracy in some cases, but tends reduce 'verbatimness' of the transcript"
                             "ignored if [suppress_silence]=False")

    parser.add_argument("--q_levels", type=int, default=20,
                        help="quantization levels for generating timestamp suppression mask; "
                             "acts as a threshold to marking sound as silent;"
                             "fewer levels will increase the threshold of volume at which to mark a sound as silent")

    parser.add_argument("--k_size", type=int, default=5,
                        help="Kernel size for average pooling waveform to generate suppression mask; "
                             "recommend 5 or 3; higher sizes will reduce detection of silence")

    parser.add_argument('--time_scale', type=float,
                        help="factor for scaling audio duration for inference;"
                             "greater than 1.0 'slows down' the audio; "
                             "less than 1.0 'speeds up' the audio; "
                             "1.0 is no scaling")

    parser.add_argument('--vad', type=str2bool, default=False,
                        help='whether to use Silero VAD to generate timestamp suppression mask; '
                             'Silero VAD requires PyTorch 1.12.0+;'
                             'Official repo: https://github.com/snakers4/silero-vad')
    parser.add_argument('--vad_threshold', type=float, default=0.35,
                        help='threshold for detecting speech with Silero VAD. (Default: 0.35); '
                             'low threshold reduces false positives for silence detection')
    parser.add_argument('--vad_onnx', type=str2bool, default=False,
                        help='whether to use ONNX for Silero VAD')

    parser.add_argument('--min_word_dur', type=float, default=get_min_word_dur(),
                        help="shortest duration each word is allowed to reach for silence suppression")
    parser.add_argument('--nonspeech_error', type=float, default=0.1,
                        help="relative error of non-speech sections that appear in between a word for "
                             "silence suppression.")

    parser.add_argument('--max_chars', type=int,
                        help="maximum number of character allowed in each segment")
    parser.add_argument('--max_words', type=int,
                        help="maximum number of words allowed in each segment")

    parser.add_argument('--demucs', type=str2bool,
                        help='whether to reprocess the audio track with Demucs to isolate vocals/remove noise; '
                             'Demucs official repo: https://github.com/facebookresearch/demucs;'
                             'DEPRECATED and replace with --denoiser "demucs"')
    parser.add_argument('--demucs_output', action="extend", nargs="+", type=str,
                        help='path(s) to save the vocals isolated by Demucs as WAV file(s); '
                             'ignored if --demucs False; DEPRECATED and replace with --denoiser_output')
    parser.add_argument('--denoiser', type=str, choices=supported_denoisers,
                        help='name of denoiser to reprocess the audio track to isolate vocals/remove noise')
    parser.add_argument('--denoiser_output', action="extend", nargs="+", type=str,
                        help='path(s) to save the denoised audio as WAV file(s); '
                             'ignored if --denoiser is unspecified')
    parser.add_argument('--only_voice_freq', '-ovf', action='store_true',
                        help='whether to only use sound between 200 - 5000 Hz, where majority of human speech are.')

    parser.add_argument('--strip', type=str2bool, default=True,
                        help="whether to remove spaces before and after text on each segment for output")

    parser.add_argument('--tag', type=str, action="extend", nargs="+",
                        help="a pair tags used to change the properties a word at its predicted time"
                             "SRT Default: '<font color=\"#00ff00\">', '</font>'"
                             "VTT Default: '<u>', '</u>'"
                             "ASS Default: '{\\1c&HFF00&}', '{\\r}'")
    parser.add_argument('--segment_level', type=str2bool, default=True,
                        help="whether to use segment-level timestamps in output")
    parser.add_argument('--word_level', type=str2bool, default=True,
                        help="whether to use word-level timestamps in output")

    parser.add_argument('--reverse_text', type=str2bool, default=False,
                        help="whether to reverse the order of words for each segment of text output")

    # ass output
    parser.add_argument('--font', type=str, default='Arial',
                        help="word font for ASS output(s)")
    parser.add_argument('--font_size', type=int, default=48,
                        help="word font size for ASS output(s)")
    parser.add_argument('--karaoke', type=str2bool, default=False,
                        help="whether to use progressive filling highlights for karaoke effect (only for ASS outputs)")

    parser.add_argument("--temperature", type=float, default=0,
                        help="temperature to use for sampling")
    parser.add_argument("--best_of", type=optional_int,
                        help="number of candidates when sampling with non-zero temperature")
    parser.add_argument("--beam_size", type=optional_int,
                        help="number of beams in beam search, only applicable when temperature is zero")
    parser.add_argument("--patience", type=float, default=None,
                        help="optional patience value to use in beam decoding, "
                             "as in https://arxiv.org/abs/2204.05424, "
                             "the default (1.0) is equivalent to conventional beam search")
    parser.add_argument("--length_penalty", type=float, default=None,
                        help="optional token length penalty coefficient (alpha) "
                             "as in https://arxiv.org/abs/1609.08144, uses simple length normalization by default")

    parser.add_argument("--suppress_tokens", type=str, default="-1",
                        help="comma-separated list of token ids to suppress during sampling; "
                             "'-1' will suppress most special characters except common punctuations")
    parser.add_argument("--initial_prompt", type=str, default=None,
                        help="optional text to provide as a prompt for the first window.")
    parser.add_argument("--condition_on_previous_text", type=str2bool, default=True,
                        help="if True, provide the previous output of the model as a prompt for the next window; "
                             "disabling may make the text inconsistent across windows, "
                             "but the model becomes less prone to getting stuck in a failure loop")
    parser.add_argument("--fp16", type=str2bool, default=True,
                        help="whether to perform inference in fp16; True by default")

    parser.add_argument("--temperature_increment_on_fallback", type=optional_float, default=0.2,
                        help="temperature to increase when falling back when the decoding fails to meet either of "
                             "the thresholds below")
    parser.add_argument("--compression_ratio_threshold", type=optional_float, default=2.4,
                        help="if the gzip compression ratio is higher than this value, treat the decoding as failed")
    parser.add_argument("--logprob_threshold", type=optional_float, default=-1.0,
                        help="if the average log probability is lower than this value, treat the decoding as failed")
    parser.add_argument("--no_speech_threshold", type=optional_float, default=0.6,
                        help="if the probability of the <|nospeech|> token is higher than this value AND the decoding "
                             "has failed due to `logprob_threshold`, consider the segment as silence")
    parser.add_argument("--threads", type=optional_int, default=0,
                        help="number of threads used by torch for CPU inference; "
                             "supercedes MKL_NUM_THREADS/OMP_NUM_THREADS")

    parser.add_argument('--mel_first', action='store_true',
                        help='process entire audio track into log-Mel spectrogram first instead in chunks; '
                             'DEPRECATED and replaced with --no_stream')
    parser.add_argument('--no_stream', action='store_true',
                        help='whether to always load the entire audio track into memory')

    parser.add_argument('--only_ffmpeg', action='store_true',
                        help='whether to use only FFmpeg (and not yt-dlp) for URls')

    parser.add_argument('--overwrite', '-y', action='store_true',
                        help='overwrite all output files')

    parser.add_argument('--debug', action='store_true',
                        help='print all input/output pair(s) and all arguments used for transcribing/translating')

    parser.add_argument('--transcribe_method', '-tm', type=str, default='transcribe',
                        choices=('transcribe', 'transcribe_minimal'))

    parser.add_argument('--align', '-a', action="extend", nargs='+', type=str,
                        help='path(s) to TXT file(s) or JSON previous result(s)')

    parser.add_argument('--refine', '-r', action='store_true',
                        help='Refine timestamps to increase precision of timestamps')

    parser.add_argument('--locate', '-lc', action="extend", nargs='+', type=str,
                        help='words to locate in the audio(s); skips transcription and output')

    parser.add_argument('--refine_option', '-ro', action="extend", nargs='+', type=str,
                        help='Extra option(s) to use for refining timestamps; Replace True/False with 1/0; '
                             'E.g. --refine_option "steps=sese" --refine_options "rel_prob_decrease=0.05"')
    parser.add_argument('--demucs_option', '-do', action="extend", nargs='+', type=str,
                        help='Extra option(s) to use for demucs; Replace True/False with 1/0; '
                             'E.g. --demucs_option "shifts=3" --demucs_options "overlap=0.5"; '
                             'DEPRECATED and replaced with --denoiser_option')
    parser.add_argument('--denoiser_option', '-dno', action="extend", nargs='+', type=str,
                        help='Extra option(s) to use for denoiser; Replace True/False with 1/0; '
                             'E.g. --denoiser_option "shifts=3" --denoiser_option "overlap=0.5"')
    parser.add_argument('--model_option', '-mo', action="extend", nargs='+', type=str,
                        help='Extra option(s) to use for loading model; Replace True/False with 1/0; '
                             'E.g. --model_option "download_root=./downloads"')
    parser.add_argument('--transcribe_option', '-to', action="extend", nargs='+', type=str,
                        help='Extra option(s) to use for transcribing/alignment/locating; Replace True/False with 1/0; '
                             'E.g. --transcribe_option "ignore_compatibility=1"')
    parser.add_argument('--save_option', '-so', action="extend", nargs='+', type=str,
                        help='Extra option(s) to use for text outputs; Replace True/False with 1/0; '
                             'E.g. --save_option "highlight_color=ffffff"')

    parser.add_argument('--faster_whisper', '-fw', action='store_true',
                        help='whether to use faster-whisper (https://github.com/guillaumekln/faster-whisper); '
                             'note: some features may not be available')
    parser.add_argument('--huggingface_whisper', '-hw', action='store_true',
                        help='whether to run Whisper on Hugging Face Transformers for more speed than faster-whisper'
                             ' and even more speed with Flash Attention enabled on supported GPUs'
                             '(https://huggingface.co/openai/whisper-large-v3); '
                             'note: some features may not be available')

    parser.add_argument('--persist', '-p', action='store_true',
                        help='Keep previous model loaded for the future sets of commands in the same CLI instance')

    args = _split_input_args(cmd) if cmd else []
    if _cache is None:
        _cache = {}
        if args:
            args = [args]
    elif _cache or args:
        if _cache and not args:
            curr_model_name = _cache['model']['fullname'] if 'model' in _cache else ''
            try:
                cmd = input(f"{curr_model_name}> ")
            except KeyboardInterrupt:
                args = []
            else:
                args = _split_input_args(cmd)

        if _cache and not args:
            _cache['persist'] = False
            return

        if _cache.get('persist') and not ('--persist' in args and '-p' in args):
            args.append('-p')

        if 'model' in _cache:
            if '--model' not in args and '-m' not in args:
                args.extend(['-m', _cache['model']['name']])
            model_type = _cache['model']['type']
            type_arg = '--faster_whisper' in args or '-fw' in args or '--huggingface_whisper' in args or '-hw' in args
            if not type_arg:
                if model_type == 'Faster-Whisper':
                    args.append('-fw')
                elif model_type == 'Hugging Face Whisper':
                    args.append('-hw')

        _, invalid_args = parser.parse_known_args(args)
        if invalid_args:
            print(f'Got invalid argument(s): {invalid_args}')
            return
        args = [args]

    args = parser.parse_args(*args).__dict__
    _cache['persist'] = args['persist']
    debug = args.pop('debug')
    if not args['language'] and (args['align'] or args['locate']):
        raise ValueError('langauge is required for --align / --locate')

    is_faster_whisper = args.pop('faster_whisper')
    is_hf_whisper = args.pop('huggingface_whisper')
    assert not (is_faster_whisper and is_hf_whisper), f'--huggingface_whisper cannot be used with --faster_whisper'
    is_original_whisper = not (is_faster_whisper or is_hf_whisper)
    model_name: str = args.pop("model")
    model_dir: str = args.pop("model_dir")
    inputs: List[Union[str, torch.Tensor]] = args.pop("inputs")
    outputs: List[str] = args.pop("output")
    output_dir: str = args.pop("output_dir")
    output_format = args.pop("output_format")
    overwrite: bool = args.pop("overwrite")
    no_stream = use_deprecated_args('no_stream', 'mel_first', pop=True, expected_default=False)
    args['stream'] = None if not no_stream else False
    if overwrite:
        set_global_overwrite_permission(True)
    denoiser = use_deprecated_args('denoiser', 'demucs', pop=True, eg='--denoiser "demucs"')
    args['denoiser'] = 'demucs' if denoiser is True else (denoiser or None)
    denoiser_outputs = use_deprecated_args('denoiser_output', 'demucs_output', pop=True)
    denoiser_options = use_deprecated_args('denoiser_option', 'demucs_option', pop=True)
    args['denoiser_options'] = update_options_with_args(denoiser_options or '') or {}
    regroup = args.pop('regroup')
    max_chars = args.pop('max_chars')
    max_words = args.pop('max_words')
    args['verbose'] = False if args['verbose'] == 1 else (True if args['verbose'] == 2 else None)
    show_curr_task = args['verbose'] is not None
    strings_to_locate = args.pop('locate')
    if dq := args.pop('dynamic_quantization', False):
        args['device'] = 'cpu'
    if args['reverse_text']:
        args['reverse_text'] = (args.get('prepend_punctuations'), args.get('append_punctuations'))

    if is_original_whisper:
        model_type_name = 'Whisper'
        from .original_whisper import load_model as load_model_func
        model_name_kwarg = dict(name=model_name)
    else:
        if is_faster_whisper:
            model_type_name = 'Faster-Whisper'
            from .faster_whisper import load_faster_whisper as load_model_func
            model_name_kwarg = dict(model_size_or_path=model_name)
        else:
            model_type_name = 'Hugging Face Whisper'
            from .hf_whisper import load_hf_whisper as load_model_func
            model_name_kwarg = dict(model_name=model_name)

        if args.get('transcribe_method') == 'transcribe_minimal':
            warnings.warn(f'{model_type_name} models already run on a version of transcribe_minimal. '
                          '--transcribe_method "transcribe_minimal" will be ignored.')
        if args.get('refine'):
            raise NotImplementedError(f'--refine is not supported for {model_type_name} models.')
        if strings_to_locate:
            raise NotImplementedError(f'--locate is not supported for {model_type_name} models.')
        if is_faster_whisper:
            args['transcribe_method'] = 'transcribe_stable'

    if regroup:
        try:
            regroup = str2bool(regroup)
        except ValueError:
            pass
    curr_output_formats: List[str] = output_format.split(',') if output_format else []
    unsupported_formats = list(set(map(str.lower, curr_output_formats)) - OUTPUT_FORMATS)
    if outputs:
        unsupported_formats.extend(list(set(splitext(o)[-1].lower().strip('.') for o in outputs) - OUTPUT_FORMATS))
    if len(unsupported_formats) != 0:
        raise NotImplementedError(f'{unsupported_formats} are not supported. Supported formats: {OUTPUT_FORMATS}.')

    if denoiser_outputs and len(denoiser_outputs) != len(inputs):
        raise ValueError(f'--denoiser_outputs and inputs do not match in count. '
                         f'Got {len(denoiser_outputs)} and {len(inputs)}')

    if tag := args.get('tag'):
        assert tag == ['-1'] or len(tag) == 2, f'[tag] must be a pair of str but got {tag}'

    def make_parent(filepath: str):
        if parent := split(filepath)[0]:
            os.makedirs(parent, exist_ok=True)

    def is_json(file: str):
        return file.endswith(".json")

    def call_method_with_options(method, options: dict, include_first: bool = True):
        def val_to_str(val) -> str:
            if isinstance(val, (np.ndarray, torch.Tensor)):
                return f'{val.__class__}(shape:{list(val.shape)})'
            elif isinstance(val, str):
                return f'"{val}"'
            elif isinstance(val, bytes):
                return f'{type(val)}(len:{len(val)})'
            elif isinstance(val, torch.nn.Module):
                return str(type(val))
            elif isinstance(val, dict):
                return str({k: val_to_str(v) for k, v in val.items()})
            return str(val)

        params = tuple(get_func_parameters(method))
        if debug:
            temp_options = {k: options.pop(k) for k in params if k in options}
            temp_options.update(options)
            options = temp_options
            options_str = ',\n'.join(
                f'    {k}={val_to_str(v)}'
                for k, v in options.items()
                if include_first or k != params[0]
            )
            if options_str:
                options_str = f'\n{options_str}\n'
            else:
                print(options, params)
            print(f'{method.__qualname__}({options_str})')
        return method(**options)

    if alignments := args['align']:
        if unsupported_align_fmts := \
                [_ext for p in alignments if (_ext := splitext(p)[-1].lower()) not in ('.json', '.txt')]:
            raise NotImplementedError(
                f'Unsupported format(s) for alignment: {unsupported_align_fmts}'
            )
        if len(inputs) != len(alignments):
            raise NotImplementedError(
                f'Got {len(inputs)} audio file(s) but specified {len(alignments)} file(s) to align.'
            )
    else:
        alignments = ['']*len(inputs)

    def finalize_outputs(input_file: str, _output: str = None, _alignment: str = None) -> List[str]:
        _curr_output_formats = curr_output_formats.copy()
        basename, ext = splitext(_output or url_to_path(input_file))
        ext = ext[1:]
        if _output:
            if ext.lower() in OUTPUT_FORMATS:
                _curr_output_formats.append(ext)
            else:
                basename = _output
        if not _curr_output_formats:
            _curr_output_formats = ["srt" if is_json(input_file) or is_json(_alignment) else "json"]
        _outputs = [f'{basename}.{ext}' for ext in set(_curr_output_formats)]
        if output_dir:
            _outputs = [join(output_dir, o) for o in _outputs]

        return _outputs

    if outputs:
        if len(outputs) != len(inputs):
            raise NotImplementedError(f'Got {len(inputs)} audio file(s) but specified {len(outputs)} output file(s).')
        final_outputs = [finalize_outputs(i, o, a) for i, o, a in zip(inputs, outputs, alignments)]
    else:
        if not output_dir:
            output_dir = '.'
        final_outputs = [finalize_outputs(i, _alignment=a) for i, a in zip(inputs, alignments)]

    if not overwrite:

        for paths in final_outputs:
            for path in paths:
                if not is_allow_overwrite(path):
                    return

    if model_name.endswith(".en") and args["language"] not in {"en", "English"}:
        if args["language"] is not None:
            warnings.warn(f"{model_name} is an English-only model but receipted "
                          f"'{args['language']}'; using English instead.")
        args["language"] = "en"

    temperature = args.pop("temperature")
    increment = args.pop("temperature_increment_on_fallback")
    if increment is not None:
        temperature = tuple(np.arange(temperature, 1.0 + 1e-6, increment))
    else:
        temperature = [temperature]

    args['temperature'] = temperature

    threads = args.pop("threads")
    if threads > 0:
        torch.set_num_threads(threads)

    if debug:
        print('Input(s)  ->  Outputs(s)')
        for i, (input_audio, output_paths, alignment) in enumerate(zip(inputs, final_outputs, alignments)):
            dm_output = f' {denoiser_outputs[i]} ->' if denoiser_outputs else ''
            alignment = f' + "{alignment}"' if alignment else ''
            print(f'"{input_audio}"{alignment}  ->{dm_output}  {output_paths}')
        print('')

    if show_curr_task:
        model_from_str = '' if model_dir is None else f' from {model_dir}'
        model_loading_str = (
            f'{model_type_name} {model_name} model {model_from_str}'
        )
        print(f'Loading {model_loading_str}\r', end='\n' if debug else '')
    else:
        model_loading_str = ''

    alignments = args['align']
    model = None

    def _load_model():
        nonlocal model
        if model is None and _cache is not None and 'model' in _cache:
            if _cache['model']['type'] == model_type_name and _cache['model']['name'] == model_name:
                model = _cache['model']['instance']
                if model_loading_str:
                    print(f"Reuse {_cache['model']['fullname'] or 'previous model'}   ")
            else:
                del _cache['model']
                import gc
                gc.collect()
        if model is None:
            model_options = dict(
                device=args.get('device'),
                download_root=model_dir,
                dq=dq,
            )
            model_options.update(model_name_kwarg)
            model_options = isolate_useful_options(model_options, load_model_func)
            update_options_with_args('model_option', model_options)
            model = call_method_with_options(load_model_func, model_options)
            if model_loading_str:
                print(f'Loaded {model_loading_str}  ')
            if _cache is not None and _cache.get('persist'):
                _cache['model'] = dict(
                    fullname=model_loading_str, name=model_name, type=model_type_name, instance=model
                )
        return model

    for i, (input_audio, output_paths) in enumerate(zip(inputs, final_outputs)):
        skip_output = False
        if isinstance(input_audio, str) and is_json(input_audio):
            result = WhisperResult(input_audio)
        else:
            model = _load_model()
            args['regroup'] = False
            args['audio'] = input_audio
            if denoiser_outputs:
                args['denoiser_options']['save_path'] = denoiser_outputs[i]
            transcribe_method = args.get('transcribe_method')
            text = None
            if alignments and (text := alignments[i]):
                if text.endswith('.json'):
                    text = WhisperResult(text)
                else:
                    with open(text, 'r', encoding='utf-8') as f:
                        text = f.read()
                args['text'] = text
                transcribe_method = 'align'
            if strings_to_locate and (text := strings_to_locate[i]):
                args['text'] = text
                transcribe_method = 'locate'
                skip_output = args['verbose'] = True
            transcribe_method = getattr(model, transcribe_method)
            transcribe_options = isolate_useful_options(args, transcribe_method)
            if not text and not is_hf_whisper:
                decoding_options = (
                    isolate_useful_options(args, model.transcribe if is_faster_whisper else DecodingOptions)
                )
                if is_faster_whisper:
                    if decoding_options['suppress_tokens']:
                        decoding_options['suppress_tokens'] = (
                            list(map(int, decoding_options['suppress_tokens'].split(',')))
                        )
                    for k in list(decoding_options.keys()):
                        if decoding_options[k] is None:
                            del decoding_options[k]
                transcribe_options.update(decoding_options)
            update_options_with_args('transcribe_option', transcribe_options)
            result: WhisperResult = call_method_with_options(transcribe_method, transcribe_options)

        if skip_output:
            continue

        if args['refine']:
            model = _load_model()
            refine_options = isolate_useful_options(args, model.refine)
            refine_options['result'] = result
            update_options_with_args('refine_option', refine_options)
            call_method_with_options(model.refine, refine_options)

        if args.get('word_timestamps'):
            if regroup:
                result.regroup(regroup, verbose=args['verbose'] or debug)
            if max_chars or max_words:
                result.split_by_length(max_chars=max_chars, max_words=max_words)

        for path in output_paths:
            make_parent(path)
            save_method = getattr(result, OUTPUT_FORMATS_METHODS[splitext(path)[-1][1:]])
            args['filepath'] = path
            args['path'] = path
            save_options = isolate_useful_options(args, save_method)
            update_options_with_args('save_option', save_options)
            call_method_with_options(save_method, save_options)


def cli(cmd: str = None):
    cache = {}
    while True:
        error = None
        try:
            _cli(cmd=cmd, _cache=cache)
        except RuntimeError as e:
            if not str(e).startswith('FFmpeg'):
                raise e
            error = e
        except ValueError as e:
            error = e
        if cache.get('persist'):
            if error is not None:
                print(f'Error: {error}')
        else:
            if error is not None:
                raise error
            break
        cmd = None
